# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wpan related functionality for both linux-ot and nrf-ot.

There are multiple implementation sharing the same CLI syntax:
 * Linux ot-ctl command
 * NRF52840 serial CLI

This module defines the common part of CLIs.
"""

import abc
from collections.abc import Collection
import contextlib
import dataclasses
import enum
import ipaddress
import itertools
import re
from typing import Any, Literal, Optional, TypedDict

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import wpan_base
from gazoo_device.utility import retry


logger = gdm_logger.get_logger()

PROPERTY_NOT_IMPLEMENTED = "Unknown: Error - unimplemented"


class Commands(str, enum.Enum):
  """OpenThread CLI commands."""

  BR_DISABLE = "br disable"
  BR_NAT64_PREFIX = "br nat64prefix"
  DATASET = "dataset"
  DATASET_CLEAR = "dataset clear"
  DATASET_COMMIT = "dataset commit {}"
  DATASET_INIT = "dataset init"
  DATASET_NETWORKKEY = "dataset networkkey"
  DATASET_SET = "dataset set"
  DIAG_START = "diag start"
  DIAG_STOP = "diag stop"
  DEVICE_LOCALE = "diag factory sysenv get nlwirelessregdom"
  FACTORY_RESET = "factoryreset"
  IFCONFIG = "ifconfig"
  IFCONFIG_UP = "ifconfig up"
  IFCONFIG_DOWN = "ifconfig down"
  IPV6_ALL_ADDRESSES = "ipaddr"
  IPV6_LINKLOCAL_ADDRESS = "ipaddr linklocal"
  IPV6_MLEID_ADDRESS = "ipaddr mleid"
  IPV6_RLOC_ADDRESS = "ipaddr rloc"
  LOCAL_OMR_PREFIX = "br omrprefix local"
  LOCAL_ONLINK_PREFIX = "br onlinkprefix local"
  MODE = "mode"
  NAT64_COUNTERS = "nat64 counters"
  NAT64_DISABLE = "nat64 disable"
  NAT64_ENABLE = "nat64 enable"
  NAT64_MAPPINGS = "nat64 mappings"
  NAT64_STATE = "nat64 state"
  NCP_CHANNEL = "channel"
  NCP_COUNTER_TX_ERR_CCA = "counters mac"
  NCP_COUNTER_TX_IP_DROPPED = "counters mac"
  NCP_COUNTER_TX_PKT_ACKED = "counters mac"
  NCP_MAC_ADDRESS = "extaddr"
  NCP_STATE = "state"
  NETDATA_SHOW = "netdata show"
  NETWORK_KEY = "networkkey"
  NETWORK_NAME = "networkname"
  NETWORK_PANID = "panid"
  NETWORK_PARTITION_ID = "partitionid"
  NETWORK_XPANID = "extpanid"
  PING = "ping"
  POLL_PERIOD = "pollperiod"
  ROUTER_SELECTION_JITTER = "routerselectionjitter"
  SCAN_ENERGY = "scan energy"
  THREAD_CHILD_TABLE = "child table"
  THREAD_NEIGHBOR_TABLE = "neighbor table"
  THREAD_NETWORK_DATA = "netdata show -x"
  THREAD_RLOC16 = "rloc16"
  THREAD_START = "thread start"
  THREAD_STOP = "thread stop"
  CSL_PERIOD = "csl period"
  CSL_TIMEOUT = "csl timeout"
  LOG_LEVEL = "log level"
  UDP_OPEN = "udp open"
  UDP_BIND = "udp bind {interface} {ipaddr} {port}"
  UDP_CLOSE = "udp close"
  UDP_SEND = "udp send {ipaddr} {port} -s {num_bytes}"


class Dataset(str, enum.Enum):
  ACTIVE = "active"
  NEW = "new"
  PENDING = "pending"


StateStr = Literal["disabled", "detached", "child", "router", "leader"]

# Parses "0 packets transmitted, 0 packets received. ...(extra info)"
_PING_STATISTICS_PATTERN = re.compile(
    r"^(?P<transmitted>\d+) packets transmitted,"
    r" (?P<received>\d+) packets received."
    r"(?: Packet loss = (?P<loss>\d+\.\d+)%.)?"
    r"(?: Round-trip min/avg/max = (?P<min>\d+)/(?P<avg>\d+\.\d+)/(?P<max>\d+)"
    r" ms.)?$"
)


class RouteFlags(str, enum.Enum):
  STABLE = "s"
  NAT64 = "n"


class Nat64StateValue(str, enum.Enum):
  UNKNOWN = "Unknown"
  DISABLED = "Disabled"
  NOT_RUNNING = "NotRunning"
  IDLE = "Idle"
  ACTIVE = "Active"


@dataclasses.dataclass(frozen=True)
class Nat64ComponentStates(TypedDict):
  """OpenThread Nat64 ComponentStates."""

  prefix_manager: Nat64StateValue
  translator: Nat64StateValue


Rloc16 = int
Prefix = tuple[ipaddress.IPv6Network, str, str, Rloc16]
Route = tuple[ipaddress.IPv6Network, frozenset[RouteFlags], str, Rloc16]
Service = tuple[int, bytes, bytes, bool, Rloc16]


class NetworkData(TypedDict):
  prefixes: list[Prefix]
  routes: list[Route]
  services: list[Service]


def _validate_hex(hexstr: str) -> None:
  """Validates that the input string in a hex string.

  Args:
    hexstr: the input string to be validated.

  Raises:
    ValueError: If the input string is not a hex string.
  """
  if len(hexstr) % 2 != 0:
    raise ValueError(f"the length of input string is not even: {hexstr}")

  for i in range(0, len(hexstr), 2):
    int(hexstr[i : i + 2], 16)


def _hex_to_bytes(hexstr: str) -> bytes:
  r"""Converts the input hex string into bytes.

  Args:
    hexstr: the input string to be converted.

  Returns:
    The result bytes.

  Example:
    Converts "36000500000e10" to b"\x36\x00\x05\x00\x00\x0e\x10"
  """
  _validate_hex(hexstr)
  return bytes(int(hexstr[i : i + 2], 16) for i in range(0, len(hexstr), 2))


def _parse_prefix(text: str) -> Prefix:
  """Parses the input text into a Thread prefix.

  Args:
    text: the input text to be parsed.

  Returns:
    The result Thread prefix

  Example:
    Parses "fdb7:5223:be73:1::/64 paos low 2000" to
    (ipaddress.IPv6Network("fdb7:5223:be73:1::/64"), "paos", "low", 0x2000)
  """
  prefix, flags, prf, rloc16 = text.removeprefix("- ").split()[:4]
  return (ipaddress.IPv6Network(prefix), flags, prf, int(rloc16, 16))


def _parse_route(text: str) -> Route:
  """Parses the input text into a Thread route.

  Args:
    text: the input text to be parsed.

  Returns:
    The result Thread route

  Example:
    Parses "fd11:4df3:4cd3:b39c::/64 s med 2000" to
    (ipaddress.IPv6Network("fd11:4df3:4cd3:b39c::/64"), True, "med", 0x2000)
  """
  line = text.split()
  if len(line) == 4:
    prefix, flag_text, prf, rloc16 = line
    # Use list to convert RouteFlags to flags list due to bug b/265366266
    flags = frozenset({flag for flag in list(RouteFlags) if flag in flag_text})
  else:
    prefix, prf, rloc16 = line
    flags: frozenset[RouteFlags] = frozenset()

  return (ipaddress.IPv6Network(prefix), flags, prf, int(rloc16, 16))


def _parse_service(text: str) -> Service:
  r"""Parses the input text into a Thread service.

  Args:
    text: the input text to be parsed.

  Returns:
    The result Thread service

  Example:
    Parses "44970 01 36000500000e10 s 2000" to
    (44970, b"\x01", b"\x36\x00\x05\x00\x00\x0e\x10", True, 0x2000)
  """
  line = text.split()

  enterprise_number, service_data, server_data = line[:3]
  if line[3] == "s":  # The third column is an optional "s" of stable flag.
    stable, rloc16 = True, line[4]
  else:
    stable, rloc16 = False, line[3]

  enterprise_number = int(enterprise_number)
  service_data = _hex_to_bytes(service_data)
  server_data = _hex_to_bytes(server_data)
  rloc16 = int(rloc16, 16)

  return (enterprise_number, service_data, server_data, stable, rloc16)


class WpanOtBase(wpan_base.WpanBase, abc.ABC):
  """Class for interacting with OpenThread CLI command and getting props."""

  @abc.abstractmethod
  def call_command(
      self, command: str, timeout: Optional[float] = None
  ) -> list[str]:
    """Calls an OpenThread CLI command and returns the response."""

  @abc.abstractmethod
  def factory_reset(self) -> None:
    """Factory resets the Thread devices."""

  @decorators.DynamicProperty
  def ncp_channel(self) -> int:
    """Gets the ncp channel."""
    return int(self.call_command(Commands.NCP_CHANNEL.value)[0])

  @decorators.DynamicProperty
  def network_is_commissioned(self) -> bool:
    """Returns whether the device is connected to a Thread network."""
    role = self.call_command(Commands.NCP_STATE.value)[0]
    return role in {"leader", "router", "child"}

  @decorators.DynamicProperty
  def ncp_mac_address(self) -> str:
    """Gets the ncp mac address."""
    return self.call_command(Commands.NCP_MAC_ADDRESS.value)[0].upper()

  @decorators.DynamicProperty
  def network_key(self) -> str:
    """Gets the Thread network master key in hex format."""
    return self.call_command(Commands.NETWORK_KEY.value)[0].upper()

  @decorators.DynamicProperty
  def network_name(self) -> str:
    """Gets the Thread network name."""
    return self.call_command(Commands.NETWORK_NAME.value)[0]

  @decorators.DynamicProperty
  def network_node_type(self) -> StateStr:
    """Gets the current node type."""
    return self.call_command(Commands.NCP_STATE.value)[0]

  @decorators.DynamicProperty
  def network_panid(self) -> str:
    """Gets the current node panid in hex format."""
    panid = int(self.call_command(Commands.NETWORK_PANID.value)[0], 16)
    return "0x{:04X}".format(panid)

  @decorators.DynamicProperty
  def network_partition_id(self) -> str:
    """Gets the Thread network partition ID.

    Returns:
      string: decimal string representation of current partition id

    Example output:
      $ ot-ctl partitionid
      577020633
      Done
    """
    return self.call_command(Commands.NETWORK_PARTITION_ID.value)[0]

  @decorators.DynamicProperty
  def ncp_rssi(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def ncp_state(self) -> StateStr:
    """Gets the current node state."""
    return self.call_command(Commands.NCP_STATE.value)[0]

  @decorators.DynamicProperty
  def network_xpanid(self) -> str:
    """Gets the extended PAN id in hex format."""
    return "0x" + self.call_command(Commands.NETWORK_XPANID.value)[0].upper()

  @decorators.DynamicProperty
  def router_selection_jitter(self) -> int:
    """Gets the router selection jitter duration in seconds."""
    return int(self.call_command(Commands.ROUTER_SELECTION_JITTER.value)[0])

  @decorators.CapabilityLogDecorator(logger)
  def set_router_selection_jitter(self, router_selection_jitter: int) -> None:
    """Sets the router selection jitter duration in seconds."""
    self.call_command(
        f"{Commands.ROUTER_SELECTION_JITTER.value} {router_selection_jitter}"
    )

  @decorators.DynamicProperty
  def thread_leader_address(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_leader_local_weight(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_leader_network_data(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_leader_router_id(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_leader_stable_network_data(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_leader_weight(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_neighbor_table(self) -> list[dict[str, str]]:
    # TODO(b/157958546) Add missing entries
    """Returns thread_neighbor table as list.

    Note:
      (1) Thread neighbor table output format version 1:
      > neighbor table
      | Role | RLOC16 | Age | Avg RSSI | Last RSSI |R|S|D|N| Extended MAC
      |
      +------+--------+-----+----------+-----------+-+-+-+-+------------------+
      |   C  | 0x6801 |   9 |      -20 |       -20 |1|1|1|1|
      d613f05fcacd04b8 |

      Done
      (2) Thread neighbor table output format version 2 (b/172644856):
      > neighbor table
      | Role | RLOC16 | Age | Avg RSSI | Last RSSI |R|D|N| Extended MAC
      |
      +------+--------+-----+----------+-----------+-+-+-+------------------+
      |   C  | 0x7801 | 136 |      -29 |       -28 |1|0|0| 5ea6be302fed82d0
      |

      Done
    """
    neighbor_table_lines = self.call_command(
        Commands.THREAD_NEIGHBOR_TABLE.value
    )
    headers = [header.strip() for header in neighbor_table_lines[0].split("|")]
    neighbor_list = []

    for row in neighbor_table_lines[2:]:
      # The last line is an empty line.
      if not row:
        break

      values = [value.strip() for value in row.split("|")]

      neighbor_entry = {
          "RLOC16": PROPERTY_NOT_IMPLEMENTED,
          "FullNetData": PROPERTY_NOT_IMPLEMENTED,
          "LinkFC": PROPERTY_NOT_IMPLEMENTED,
          "LQIn": PROPERTY_NOT_IMPLEMENTED,
          "Age": PROPERTY_NOT_IMPLEMENTED,
          "RxOnIdle": PROPERTY_NOT_IMPLEMENTED,
          "Id": PROPERTY_NOT_IMPLEMENTED,
          "FTD": PROPERTY_NOT_IMPLEMENTED,
          "LastRssi": PROPERTY_NOT_IMPLEMENTED,
          "MleFC": PROPERTY_NOT_IMPLEMENTED,
          "SecDataReq": PROPERTY_NOT_IMPLEMENTED,
          "AveRssi": PROPERTY_NOT_IMPLEMENTED,
          "IsChild": PROPERTY_NOT_IMPLEMENTED,
      }

      for key, value in zip(headers, values):
        if key == "Role":
          neighbor_entry["IsChild"] = "yes" if value == "C" else "no"
        elif key == "RLOC16":
          neighbor_entry["RLOC16"] = "{:04x}".format(
              int(value, 16)
          )  # Remove the 0x
        elif key == "Age":
          neighbor_entry["Age"] = value
        elif key == "Avg RSSI":
          neighbor_entry["AveRssi"] = value
        elif key == "Last RSSI":
          neighbor_entry["LastRssi"] = value
        elif key == "R":
          neighbor_entry["RxOnIdle"] = "yes" if value == "1" else "no"
        elif key == "S":
          neighbor_entry["SecDataReq"] = "yes" if value == "1" else "no"
        elif key == "D":
          neighbor_entry["FTD"] = "yes" if value == "1" else "no"
        elif key == "N":
          neighbor_entry["FullNetData"] = "yes" if value == "1" else "no"
        elif key == "Extended MAC":
          neighbor_entry["Id"] = value.upper()

      neighbor_list.append(neighbor_entry)

    return neighbor_list

  @decorators.DynamicProperty
  def thread_network_data(self) -> str:
    """Returns the network data in the hex format.

    Example output:
      # ot-ctl netdata show -x
      08040b024d5c0b0e8001010d090c0068000500000e1003140040fd49f7e80076000105040c00f10007021140030f0040fd65aa133519275f01030c00000b1981015d0d140c00fdd2cb0d665f000023a57c9f08424339d11f
      Done
    """
    response = self.call_command(Commands.THREAD_NETWORK_DATA.value)
    if response:
      return response[0]
    else:
      return ""

  @decorators.DynamicProperty
  def thread_rloc16(self) -> str:
    """Gets the node's routing locator in hex format."""
    return "0x" + self.call_command(Commands.THREAD_RLOC16.value)[0].upper()

  @decorators.DynamicProperty
  def thread_router_id(self) -> str:
    """Gets the node's router id in hex format."""
    rloc = int(self.thread_rloc16, 16)
    # rloc's uppper byte is router id
    return "0x{:02X}".format(rloc >> 8)

  @decorators.DynamicProperty
  def thread_stable_network_data(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def thread_stable_network_data_version(self) -> str:
    # TODO(b/157958546)
    return PROPERTY_NOT_IMPLEMENTED

  @decorators.DynamicProperty
  def ncp_counter_tx_err_cca(self) -> int:
    """Gets the number of sent packets that failed with cca error."""
    counters = self.call_command(Commands.NCP_COUNTER_TX_ERR_CCA.value)
    acked = next(c for c in counters if c.find("TxErrCca") >= 0)
    return int(acked.split()[-1])

  @decorators.DynamicProperty
  def ncp_counter_tx_ip_dropped(self) -> int:
    """Gets the number of dropped sent packets."""
    counters = self.call_command(Commands.NCP_COUNTER_TX_IP_DROPPED.value)
    acked = next(c for c in counters if c.find("TxErrBusyChannel") >= 0)
    return int(acked.split()[-1])

  @decorators.DynamicProperty
  def ncp_counter_tx_pkt_acked(self) -> int:
    """Gets the number of acknowledged sent packets."""
    counters = self.call_command(Commands.NCP_COUNTER_TX_PKT_ACKED.value)
    acked = next(c for c in counters if c.find("TxAcked") >= 0)
    return int(acked.split()[-1])

  @decorators.DynamicProperty
  def thread_child_table(self) -> list[dict[str, str]]:
    """Returns thread_child_table as list of lines.

    Note:
      (1) Thread child table format version 1:

      > child table
      | ID  | RLOC16 | Timeout    | Age        | LQ In | C_VN |R|S|D|N|
      Extended MAC     |
      +-----+--------+------------+------------+-------+------+-+-+-+-+------------------+
      |   1 | 0x5c01 |        240 |         11 |     3 |  127 |1|1|0|0|
      521ae942170d88ed |

      Done

      (2) Thread child table format version 2 (b/172644856):
      > child table
      | ID  | RLOC16 | Timeout    | Age        | LQ In | C_VN |R|D|N|
      Extended MAC     |
      +-----+--------+------------+------------+-------+------+-+-+-+------------------+
      |   1 | 0x1c01 |        240 |         17 |     3 |   68 |1|0|0|
      6ae0664fdc316f66 |

      Done

      (3) Thread child table format version 3 (b/172644856):
      > child table
      | ID  | RLOC16 | Timeout    | Age        | LQ In | C_VN
      |R|D|N|Ver|CSL|QMsgCnt|
      +-----+--------+------------+------------+-------+------+-+-+-+---+---+-------+
      |   1 | 0xb001 |        240 |         12 |     3 |  168 |1|0|0|  2| 0
      |     0 |

      Done

       Extended MAC     |
      ------------------+
       9ab3267c95e504b9 |
    """
    # TODO(b/157958546) Add missing entries
    child_table_lines = self.call_command(Commands.THREAD_CHILD_TABLE.value)
    headers = [header.strip() for header in child_table_lines[0].split("|")]
    children_list = []

    for row in child_table_lines[2:]:
      # The last line is an empty line.
      if not row:
        break

      values = [value.strip() for value in row.split("|")]

      child_entry = {
          "RLOC16": PROPERTY_NOT_IMPLEMENTED,
          "FullNetData": PROPERTY_NOT_IMPLEMENTED,
          "NetDataVer": PROPERTY_NOT_IMPLEMENTED,
          "LQIn": PROPERTY_NOT_IMPLEMENTED,
          "Age": PROPERTY_NOT_IMPLEMENTED,
          "RxOnIdle": PROPERTY_NOT_IMPLEMENTED,
          "FTD": PROPERTY_NOT_IMPLEMENTED,
          "LastRssi": PROPERTY_NOT_IMPLEMENTED,
          "Timeout": PROPERTY_NOT_IMPLEMENTED,
          "SecDataReq": PROPERTY_NOT_IMPLEMENTED,
          "AveRssi": PROPERTY_NOT_IMPLEMENTED,
          "Id": PROPERTY_NOT_IMPLEMENTED,
      }

      for key, value in zip(headers, values):
        if key == "RLOC16":
          child_entry["RLOC16"] = "{:04x}".format(
              int(value, 16)
          )  # Remove the 0x
        elif key == "Timeout":
          child_entry["Timeout"] = value
        elif key == "Age":
          child_entry["Age"] = value
        elif key == "LQ In":
          child_entry["LQIn"] = value
        elif key == "C_VN":
          child_entry["NetDataVer"] = value
        elif key == "R":
          child_entry["RxOnIdle"] = "yes" if value == "1" else "no"
        elif key == "S":
          child_entry["SecDataReq"] = "yes" if value == "1" else "no"
        elif key == "D":
          child_entry["FTD"] = "yes" if value == "1" else "no"
        elif key == "N":
          child_entry["FullNetData"] = "yes" if value == "1" else "no"
        elif key == "Extended MAC":
          child_entry["Id"] = value.upper()

      children_list.append(child_entry)

    return children_list

  @decorators.DynamicProperty
  def ipv6_all_addresses(self) -> list[str]:
    """Returns ipv6:AllAddresses as list of lines."""
    return self.call_command(Commands.IPV6_ALL_ADDRESSES.value)

  @decorators.DynamicProperty
  def scan_energy(self) -> dict[int, int]:
    """Returns a dict from channel to RSSI.

    Note:
      % ot-ctl scan energy
      returns a table like

      | Ch | RSSI |
      +----+------+
      | 11 |  -42 |
      | 12 |  -84 |
      | 13 |  -86 |
      | 14 |  -89 |

      Will be converted to:

      {
        11: -42,
        12: -84,
        13: -86,
        14: -89
      }
    """
    lines = self.call_command(Commands.SCAN_ENERGY.value)

    def line_to_tuple(line: str):
      data = line.split("|")
      return int(data[1]), int(data[2])

    return dict(map(line_to_tuple, lines[2:]))

  @decorators.CapabilityLogDecorator(logger)
  def dataset_init(self, dataset: Dataset) -> None:
    """Initializes operational dataset."""
    self.call_command(f"{Commands.DATASET_INIT.value} {dataset.value}")

  @decorators.CapabilityLogDecorator(logger)
  def dataset_clear(self) -> None:
    """Clears operational dataset buffer."""
    self.call_command(Commands.DATASET_CLEAR.value)

  @decorators.CapabilityLogDecorator(logger)
  def dataset_commit(self, dataset: Dataset) -> None:
    """Commits operational dataset."""
    if dataset in (Dataset.ACTIVE, Dataset.PENDING):
      self.call_command(Commands.DATASET_COMMIT.value.format(dataset.value))
    else:
      raise ValueError(f"{self._device_name} invalid dataset: {dataset}.")

  def get_mode(self) -> str:
    """Gets the Thread Device Mode value.

    Returns:
      A string containing combination of following characters.
        "-": no flags set.
        "r": rx-on-when-idle.
        "d": Full Thread Device.
        "n": Full Network Data.
    """
    return self.call_command(Commands.MODE.value)[0]

  @decorators.CapabilityLogDecorator(logger)
  def set_mode(self, mode: str) -> None:
    """Sets the Thread Device Mode value.

    Args:
      mode: combination of following characters. "-": no flags set. "r":
        rx-on-when-idle. "d": Full Thread Device. "n": Full Network Data.
    """
    self.call_command(f"{Commands.MODE.value} {mode}")

  @decorators.CapabilityLogDecorator(logger)
  def ifconfig_up(self) -> None:
    """Bring up the IPv6 interface."""
    self.call_command(Commands.IFCONFIG_UP.value)

  @decorators.CapabilityLogDecorator(logger)
  def ifconfig_down(self) -> None:
    """Bring down the IPv6 interface."""
    self.call_command(Commands.IFCONFIG_DOWN.value)

  def get_ifconfig_state(self) -> bool:
    """Gets the status of the IPv6 interface."""
    result = self.call_command(Commands.IFCONFIG.value)[-1]
    if result not in ("up", "down"):
      raise errors.DeviceError(
          f"{self._device_name} Malformed ifconfig return value {result}"
      )

    return result == "up"

  @decorators.CapabilityLogDecorator(logger)
  def thread_start(self) -> None:
    """Enable Thread protocol operation and attach to a Thread network."""
    self.call_command(Commands.THREAD_START.value)

  @decorators.CapabilityLogDecorator(logger)
  def thread_stop(self) -> None:
    """Disable Thread protocol operation and detach from a Thread network."""
    self.call_command(Commands.THREAD_STOP.value)

  @decorators.CapabilityLogDecorator(logger)
  def get_dataset_string(
      self, dataset: Literal[Dataset.ACTIVE, Dataset.PENDING]
  ) -> str:
    return self.call_command(f"{Commands.DATASET.value} {dataset.value} -x")[0]

  @decorators.CapabilityLogDecorator(logger)
  def set_dataset_string(
      self, dataset: Literal[Dataset.ACTIVE, Dataset.PENDING], data: str
  ) -> None:
    self.call_command(f"{Commands.DATASET_SET.value} {dataset.value} {data}")

  @decorators.DynamicProperty
  def network_data(self) -> NetworkData:
    r"""Returns the network data.

    Example output:
      {
        "prefixes": [(
          ipaddress.IPv6Network("fdb7:5223:be73:1::/64"), "paos", "low", 0x2000
        )],
        "routes": [(
          ipaddress.IPv6Network("fd11:4df3:4cd3:b39c::/64"), True, "med", 0x2000
        )],
        "services": [
          (44970, b"\x01", b"\x36\x00\x05\x00\x00\x0e\x10", True, 0x2000),
        ],
      }
    """
    output = self.call_command(Commands.NETDATA_SHOW.value)
    iterator = iter(output)

    netdata: NetworkData = {
        "prefixes": [],
        "routes": [],
        "services": [],
    }

    if next(iterator) != "Prefixes:":
      raise errors.DeviceError(
          f"{self._device_name} Malformed network_data result {output}"
      )

    netdata["prefixes"] = [
        _parse_prefix(line)
        for line in itertools.takewhile(lambda v: v != "Routes:", iterator)
    ]

    netdata["routes"] = [
        _parse_route(line)
        for line in itertools.takewhile(lambda v: v != "Services:", iterator)
    ]

    netdata["services"] = [_parse_service(line) for line in iterator]

    return netdata

  @decorators.DynamicProperty
  def local_omr_prefix(self) -> str:
    """Returns the local OMR (Off-Mesh-Routable) prefix."""
    return self.call_command(Commands.LOCAL_OMR_PREFIX.value)[0].split(" ")[0]

  @decorators.DynamicProperty
  def local_onlink_prefix(self) -> str:
    """Returns the local On-Link prefix."""
    return self.call_command(Commands.LOCAL_ONLINK_PREFIX.value)[0]

  @decorators.DynamicProperty
  def network_data_omr_prefixes(self) -> list[Prefix]:
    """Returns the OMR (Off-Mesh-Routable) prefixes in network data.

    Most of the time the list should only have 1 entry.

    Example output:
      [(
          ipaddress.IPv6Network("fdb7:5223:be73:1::/64"), "paos", "med", 0x2000
      ),]
    """
    prefixes: list[Prefix] = self.network_data["prefixes"]
    result: list[Prefix] = []

    for prefix in prefixes:
      _, flags, _, _ = prefix
      if (
          "s" in flags
          and "o" in flags
          and "a" in flags
          and "d" not in flags
          and "N" not in flags
          and "D" not in flags
      ):
        result.append(prefix)

    return result

  @decorators.DynamicProperty
  def omr_addresses(self) -> list[str]:
    """Returns the OMR (Off-Mesh-Routable) addresses of this device.

    Most of the time the list should only have 1 entry.

    Example output:
      ["fd90:573a:2fcc:1239:2a44:f384:c14d:bb59",]
    """
    omr_prefixes: list[Prefix] = self.network_data_omr_prefixes
    result: list[str] = []

    for address in self.ipv6_all_addresses:
      if any(
          [
              ipaddress.ip_address(address) in prefix[0]
              for prefix in omr_prefixes
          ]
      ):
        result.append(address)

    return result

  @decorators.CapabilityLogDecorator(logger)
  def wait_for_state(
      self, states: Collection[StateStr], timeout: float = 30
  ) -> None:
    """Waits for the board going to the expected state.

    Args:
      states: A set of expected states chosen from "disabled", "detached",
        "child", "router", "leader"
      timeout: Max time to wait
    """
    retry.retry(
        func=lambda: self.ncp_state,
        is_successful=lambda state: state in states,
        reraise=False,
        timeout=timeout,
        interval=1,
    )

  @decorators.DynamicProperty
  def active_dataset(self) -> str:
    """Returns the active dataset in hex format.

    Example output:
      # ot-ctl dataset active -x
      0e080000000000010000000300001335060004001fffc00208a4b829ec0e6d24cd0708fda2c4980319febe051092a727e1abd0ed40f50c548a2b5b7edd030f4f70656e5468726561642d3462316201024b1b0410881d7f8c09a3d9d72b9540152fc32ff60c0402a0f7f8
      Done
    """
    return self.get_dataset_string(Dataset.ACTIVE)

  def _parse_ping_result(self, result: str) -> dict[str, Any]:
    """Parses ping result."""
    statistics = {}
    match = _PING_STATISTICS_PATTERN.match(result)
    if match is not None:
      if match.group("transmitted") is not None:
        statistics["transmitted_packets"] = int(match.group("transmitted"))
        statistics["received_packets"] = int(match.group("received"))
      if match.group("loss") is not None:
        statistics["packet_loss"] = float(match.group("loss")) / 100
      if match.group("min") is not None:
        statistics["round_trip_time"] = {
            "min": int(match.group("min")),
            "avg": float(match.group("avg")),
            "max": int(match.group("max")),
        }
    else:
      raise errors.DeviceError(
          f"{self._device_name} Malformed ping result {result}"
      )
    return statistics

  def ping(
      self,
      ipaddr: str,
      size: int = 60,
      count: int = 1,
      interval: float = 1.0,
      hoplimit: int = 64,
      timeout: float = 3.0,
  ) -> dict[str, Any]:
    """Runs ping command on the board.

    Args:
      ipaddr: destination to ping.
      size: ping packet size.
      count: number of ping packets.
      interval: interval between sending ping request.
      hoplimit: TTL in the ping packet.
      timeout: total timeout for the ping command.

    Returns:
      A dict of ping result. Example:
        {
          "transmitted_packets": 1,
          "received_packets": 1,
          "packet_loss": 0.0,
          "round_trip_time": {
            "min": 2,
            "avg": 2.0,
            "max": 2
          }
        }
    """

    timeout_allowance = 3
    cli_timeout = (count - 1) * interval + timeout + timeout_allowance
    cmd = (
        f"{Commands.PING.value} {ipaddr} {size} {count} {interval} {hoplimit} {timeout}"
    )

    result = self.call_command(cmd, timeout=cli_timeout)[-1]
    return self._parse_ping_result(result)

  @decorators.DynamicProperty
  def ipaddr_mleid(self) -> str:
    """Get Thread Mesh Local EID address."""
    return self.call_command(Commands.IPV6_MLEID_ADDRESS.value)[0]

  @decorators.DynamicProperty
  def ipaddr_linklocal(self) -> str:
    """Get Thread link-local IPv6 address."""
    return self.call_command(Commands.IPV6_LINKLOCAL_ADDRESS.value)[0]

  @decorators.DynamicProperty
  def ipaddr_rloc(self) -> str:
    """Get Thread Routing Locator (RLOC) address."""
    return self.call_command(Commands.IPV6_RLOC_ADDRESS.value)[0]

  @decorators.CapabilityLogDecorator(logger)
  def disable_border_routing(self) -> None:
    """Disables border routing."""
    self.call_command(Commands.BR_DISABLE.value)

  @decorators.DynamicProperty
  def poll_period(self) -> int:
    """Gets the customized data poll period for the device.

    Returns:
      The data poll period in milliseconds.
    """
    return int(self.call_command(Commands.POLL_PERIOD.value)[0])

  @decorators.CapabilityLogDecorator(logger)
  def set_poll_period(self, poll_period: int) -> None:
    """Sets the customized data poll period for the device.

    Args:
      poll_period: The data poll period in milliseconds.
    """
    self.call_command(f"{Commands.POLL_PERIOD.value} {poll_period}")

  @decorators.DynamicProperty
  def log_level(self) -> int:
    """Gets the Thread log level for the device.

    Returns:
      The Thread log level (from 1-5).
    """
    return int(self.call_command(Commands.LOG_LEVEL.value)[0])

  @decorators.CapabilityLogDecorator(logger)
  def set_log_level(self, log_level: int) -> None:
    """Sets the Thread log level for the device.

    Args:
      log_level: The log level (from 1-5).
    """
    self.call_command(f"{Commands.LOG_LEVEL.value} {log_level}")

  @decorators.CapabilityLogDecorator(logger)
  def get_device_locale(self) -> list[str]:
    """Returns the Device Locale information.

    Note:
      get_device_locale can only fetch in right after a factory reset, it will
      raise an InvalidState error when called in a wrong state.
    """
    self.call_command(Commands.DIAG_START.value)
    device_locale_info = self.call_command(Commands.DEVICE_LOCALE.value)
    self.call_command(Commands.DIAG_STOP.value)
    return device_locale_info

  @decorators.CapabilityLogDecorator(logger)
  def udp_open(self) -> None:
    """Opens the UDP socket."""
    self.call_command(Commands.UDP_OPEN.value)

  @decorators.CapabilityLogDecorator(logger)
  def udp_bind(
      self, ipaddr: str, port: int, bind_unspecified: bool = False
  ) -> None:
    """Binds the UDP socket to the specific address and port.

    Args:
      ipaddr: The local address to bind.
      port: The local port to bind.
      bind_unspecified: Whether to bind all interfaces on the device. False for
        binding to Thread interface only, True for binding to all interfaces.
    """
    # TODO(b/269228055): The OpenThread CLI does not return the actual port
    # bound. Return a pair of (address, port) after OpenThread CLI prints the
    # address and port for `udp bind`.
    self.call_command(
        Commands.UDP_BIND.value.format(
            interface="-u" if bind_unspecified else "",
            ipaddr=ipaddr,
            port=port,
        )
    )

  @decorators.CapabilityLogDecorator(logger)
  def udp_close(self) -> None:
    """Closes the UDP socket."""
    self.call_command(Commands.UDP_CLOSE.value)

  @decorators.CapabilityLogDecorator(logger)
  def udp_send(self, num_bytes: int, ipaddr: str, port: int) -> None:
    """Sends a udp packet to the given address and port.

    Args:
      num_bytes: The length of generated UDP payload.
      ipaddr: The destination address.
      port: The destination port.
    """
    self.call_command(
        Commands.UDP_SEND.value.format(
            ipaddr=ipaddr, port=port, num_bytes=num_bytes
        )
    )

  @contextlib.contextmanager
  @decorators.CapabilityLogDecorator(logger)
  def open_udp_and_bind(
      self, ipaddr: str, port: int, bind_unspecified: bool = False
  ):
    """Binds the UDP socket to the specific address and port.

    This function is expected to work with `with` statement. The UDP socket on
    device will be closed automatically.

    ```
    with device.wpan.open_udp_and_bind("::", 12345):
      # Do something
      # device.wpan.udp_close() will be called automatically.
    ```

    Args:
      ipaddr: The local address to bind.
      port: The local port to bind.
      bind_unspecified: Whether to bind all interfaces on the device. False for
        binding to Thread interface only, True for binding to all interfaces.

    Yields:
      `None`
    """
    self.udp_open()
    try:
      yield self.udp_bind(ipaddr, port, bind_unspecified)
    finally:
      self.udp_close()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RPC implementation of Matter Boolean State cluster capability.
"""

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import boolean_state_base
from gazoo_device.protos import attributes_service_pb2


logger = gdm_logger.get_logger()
_BooleanStateCluster = matter_enums.BooleanStateCluster
BOOLEAN_ATTRIBUTE_TYPE = attributes_service_pb2.AttributeType.ZCL_BOOLEAN_ATTRIBUTE_TYPE


class BooleanStateClusterPwRpc(
    boolean_state_base.BooleanStateClusterBase):
  """Matter Boolean State cluster capability."""

  @decorators.DynamicProperty
  def state_value(self) -> bool:
    """The StateValue attribute.

    The semantics of this boolean state are defined by the device type using
    this cluster. For example, in a Contact Sensor device type, FALSE=open or
    no contact, TRUE=closed or contact.

    Returns:
      The StateValue attribute.
    """
    measured_value_data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=_BooleanStateCluster.ID,
        attribute_id=_BooleanStateCluster.ATTRIBUTE_STATE_VALUE,
        attribute_type=BOOLEAN_ATTRIBUTE_TYPE)
    return measured_value_data.data_bool

  @state_value.setter
  def state_value(self, state: bool) -> None:
    """Updates the StateValue attribute.

    The semantics of this boolean state are defined by the device type using
    this cluster. For example, in a Contact Sensor device type, FALSE=open or
    no contact, TRUE=closed or contact.

    Args:
      state: The state BooleanState should update to.
    """
    previous_state = self.state_value

    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=_BooleanStateCluster.ID,
        attribute_id=_BooleanStateCluster.ATTRIBUTE_STATE_VALUE,
        attribute_type=BOOLEAN_ATTRIBUTE_TYPE,
        data_bool=state)

    if self.state_value != state:  # pylint: disable=comparison-with-callable
      raise errors.DeviceError(
          f"Device {self._device_name} state_value didn't change to "
          f"{state} from {previous_state}.")




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC transport class."""
import fcntl
import importlib
import queue
import select
import socket
import threading
import typing
from typing import Any, Collection, Optional, Union
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.switchboard.transports import transport_base
from gazoo_device.utility import pwrpc_utils
import serial

# pylint: disable=g-import-not-at-top
# pytype: disable=import-error
try:
  raise ImportError
except ImportError:
  from pw_rpc import client
  from pw_rpc import callback_client
  from pw_hdlc import rpc
  from pw_hdlc import decode
  from pw_protobuf_compiler import python_protos
# pylint: enable=g-import-not-at-top
# pytype: enable=import-error

_STDOUT_ADDRESS = 1
_DEFAULT_ADDRESS = ord("R")
_JOIN_TIMEOUT_SEC = 1  # seconds
_SELECT_TIMEOUT_SEC = 0.1  # seconds
_SERIAL_TIMEOUT_SEC = 0.01  # seconds
_RPC_CALLBACK_TIMEOUT_SEC = 1  # seconds
_NUM_OF_READ_BYTES = 4096
RPC_METHOD_NAME = "rpc"  # Pigweed RPC method name for switchboard call
logger = gdm_logger.get_logger()


def _is_primitive(var: Any) -> bool:
  """Checks if the given variable is of a primitive type."""
  primitive = (bool, bytes, float, int, str)
  return isinstance(var, primitive)


def _serialize(payload: Any) -> Any:
  """Serializes the given payload.

  Args:
    payload: The payload to serialize. Note that the payload type will be either
      primitive type, protobuf type, or the list type of the above combination.

  Returns:
    Serialized payload.
  """
  if _is_primitive(payload):
    return payload
  if not isinstance(payload, list):
    return payload.SerializeToString()
  return [_serialize(content) for content in payload]


class PwHdlcRpcClient:
  """Pigweed HDLC RPC Client.

  The RPC client class for sending RPC requests and receiving RPC responses.
  The client supports HDLC communication over serial and socket connection.

  Adapted from
  https://pigweed.googlesource.com/pigweed/pigweed/+/refs/heads/master/pw_hdlc/py/pw_hdlc/rpc.py.
  """

  def __init__(self,
               file_object: Union[serial.Serial, socket.socket],
               protobuf_import_paths: Collection[str]):
    """Creates an RPC client configured to communicate using HDLC.

    Args:
      file_object: File descriptor object to interact, can be serial interface
        instance or socket interface instance.
      protobuf_import_paths: Import paths of the RPC proto modules.
    """
    self._file_object = file_object
    self._protobuf_import_paths = protobuf_import_paths
    self._client = None
    self._stop_event = None
    self._worker = None
    self.log_queue = None

    # The read / write methods for 2 types of file descriptors
    if isinstance(self._file_object, serial.Serial):
      self._file_object = typing.cast(serial.Serial, self._file_object)
      self._write_method = self._file_object.write
      self._read_method = self._file_object.read
    elif isinstance(self._file_object, socket.socket):
      self._file_object = typing.cast(socket.socket, self._file_object)
      self._write_method = self._file_object.sendall
      self._read_method = self._file_object.recv
    else:
      raise errors.DeviceError(
          "Invalid file object type. Must be serial.Serial or "
          f"socket.socket. The given object is {type(self._file_object)}.")

  def is_alive(self) -> bool:
    """Returns true if the worker thread has started."""
    return self._worker is not None and self._worker.is_alive()

  def start(self) -> None:
    """Creates and starts the worker thread if it hasn't been created."""
    if self._worker is None:
      self._stop_event = threading.Event()
      self.log_queue = queue.Queue()
      proto_modules = [importlib.import_module(import_path)
                       for import_path in self._protobuf_import_paths]
      # Load compiled proto modules instead of the raw protos to avoid a
      # "protoc" dependency.
      protos = python_protos.Library.from_paths(proto_modules)
      client_impl = callback_client.Impl(
          default_unary_timeout_s=_RPC_CALLBACK_TIMEOUT_SEC,
          default_stream_timeout_s=_RPC_CALLBACK_TIMEOUT_SEC)
      channels = rpc.default_channels(self._write_method)
      self._client = client.Client.from_modules(client_impl,
                                                channels,
                                                protos.modules())
      self._worker = threading.Thread(target=self.read_and_process_data)
      self._worker.start()

  def close(self) -> None:
    """Sets the threading event and joins the worker thread."""
    if self._worker is not None:
      self._stop_event.set()
      self._worker.join(timeout=_JOIN_TIMEOUT_SEC)
      if self._worker.is_alive():
        raise errors.DeviceError(
            f"The child thread failed to join after {_JOIN_TIMEOUT_SEC} seconds"
            )
      self._worker = None
      self._client = None
      self._stop_event = None
      self.log_queue = None

  def rpcs(self, channel_id: Optional[int] = None) -> Any:
    """Returns object for accessing services on the specified channel.

    Args:
      channel_id: None or RPC channel id.

    Returns:
      RPC instance over the specificed channel.
    """
    if self._client is None:
      return None
    if channel_id is None:
      return next(iter(self._client.channels())).rpcs
    else:
      return self._client.channel(channel_id).rpcs

  def read_and_process_data(self) -> None:
    """Continuously reads and handles HDLC frames."""
    decoder = decode.FrameDecoder()
    while not self._stop_event.is_set():
      readfds, _, _ = select.select(
          [self._file_object], [], [], _SELECT_TIMEOUT_SEC)
      if readfds:
        try:
          data = self._read_method(_NUM_OF_READ_BYTES)
        except Exception:  # pylint: disable=broad-except
          logger.exception(
              "Exception occurred when reading in PwHdlcRpcClient thread.")
          data = None
        if data:
          for frame in decoder.process_valid_frames(data):
            self._handle_frame(frame)

  def _handle_rpc_packet(self, frame: Any) -> None:
    """Handler for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    if not self._client.process_packet(frame.data):
      logger.error(f"Packet not handled by RPC client: {frame.data}")

  def _push_to_log_queue(self, frame: Any) -> None:
    """Pushes the HDLC log in frame into the log queue.

    Args:
      frame: HDLC frame packet.
    """
    self.log_queue.put(frame.data + b"\n")

  def _handle_frame(self, frame: Any) -> None:
    """Private method for processing HDLC frame.

    Args:
      frame: HDLC frame packet.
    """
    try:
      if not frame.ok():
        return
      if frame.address == _DEFAULT_ADDRESS:
        self._handle_rpc_packet(frame)
      elif frame.address == _STDOUT_ADDRESS:
        self._push_to_log_queue(frame)
      else:
        logger.warning(f"Unhandled frame for address {frame.address}: {frame}")
    except Exception:  # pylint: disable=broad-except
      logger.exception("Exception occurred in PwHdlcRpcClient.")


def _rpc(hdlc_client: PwHdlcRpcClient,
         service_name: str,
         event_name: str,
         **kwargs: Any) -> bytes:
  """RPC call to the Matter endpoint with given service and event name.

  Args:
    hdlc_client: HDLC client instance.
    service_name: PwRPC service name.
    event_name: Event name in the given service instance.
    **kwargs: Arguments for the event method.

  Returns:
    RPC encoded payload in bytes.

  Raises:
    DeviceError when the HDLC client is not alive or RPC ack value is not OK.
  """
  if not hdlc_client.is_alive():
    raise errors.DeviceError("HLDC client is not alive.")
  client_channel = hdlc_client.rpcs().chip.rpc
  service = getattr(client_channel, service_name)
  event = getattr(service, event_name)
  kwargs = {
      param_name:
      param.decode() if isinstance(param, pwrpc_utils.PigweedProtoState)
      else param for param_name, param in kwargs.items()}
  ack, payload = event(**kwargs)
  if not ack.ok():
    raise errors.DeviceError(
        f"Pigweed RPC call {service_name} {event_name} {kwargs} fails: "
        f"Error message: {ack.name}. Error code: {ack.value}.")
  return _serialize(payload)


class PigweedRpcSerialTransport(transport_base.TransportBase):
  """Pigweed RPC transport over serial connection via UART."""

  def __init__(self,
               comms_address: str,
               protobuf_import_paths: Collection[str],
               baudrate: int,
               auto_reopen: bool = True,
               open_on_start: bool = True):
    """Initializes a PigweedRpcSerialTransport instance.

    Args:
      comms_address: Serial port path on the host.
      protobuf_import_paths: Module import paths of the compiled device
        communication protobufs. For example,
        ["gazoo_device.protos.device_service_pb2"].
        Proto module objects are not serializable, so their import paths are
        stored instead. Transport processes import the proto modules from these
        paths.
      baudrate: Serial baud rate.
      auto_reopen: Whether to automatically reopen the transport if it closes
        unexpectedly.
      open_on_start: Whether to open the transport during TransportProcess
        start.
    """
    super().__init__(
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)
    self.comms_address = comms_address
    self._serial = serial.Serial()
    self._serial.port = comms_address
    self._serial.baudrate = baudrate
    self._serial.timeout = _SERIAL_TIMEOUT_SEC
    self._hdlc_client = PwHdlcRpcClient(self._serial, protobuf_import_paths)

  def is_open(self) -> bool:
    """Returns True if the PwRPC transport is connected to the target.

    Returns:
      True if transport is open, False otherwise.
    """
    return self._serial.isOpen() and self._hdlc_client.is_alive()

  def _close(self) -> None:
    """Closes the PwRPC transport."""
    self._hdlc_client.close()
    fcntl.flock(self._serial.fileno(), fcntl.LOCK_UN)
    self._serial.close()

  def _open(self) -> None:
    """Opens the PwRPC transport."""
    self._serial.open()
    fd = self._serial.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFD)
    fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
    self._hdlc_client.start()

  def _read(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, size: int, timeout: float) -> bytes:
    """Returns Pigweed log from the HDLC channel 1.

    Args:
      size: Not used.
      timeout: Maximum seconds to wait to read bytes.

    Returns:
      bytes read from transport or None if no bytes were read
    """
    # Retrieving logs from queue doesn't support size configuration.
    del size  # not used
    try:
      return self._hdlc_client.log_queue.get(timeout=timeout)
    except queue.Empty:
      return b""

  def _write(self, data: str, timeout: Optional[float] = None) -> int:
    """Dummy method for the abstract parent class."""
    del data  # not used
    del timeout  # not used
    return 0

  def rpc(self,
          service_name: str,
          event_name: str,
          **kwargs: Any) -> bytes:
    """RPC call to the Matter endpoint with given service and event name."""
    return _rpc(self._hdlc_client, service_name, event_name, **kwargs)


class PigweedRpcSocketTransport(transport_base.TransportBase):
  """Pigweed RPC Transport over socket connection."""

  def __init__(self,
               comms_address: str,
               protobuf_import_paths: Collection[str],
               port: int,
               auto_reopen: bool = True,
               open_on_start: bool = False):
    """Initializes a PigweedRpcSocketTransport instance.

    Args:
      comms_address: Serial port path on the host.
      protobuf_import_paths: Module import paths of the compiled device
        communication protobufs.
      port: Socket connection port.
      auto_reopen: Whether to automatically reopen the transport if it closes
        unexpectedly.
      open_on_start: Whether to open the transport during TransportProcess
        start. Set to False as the transport should be opened only if the linux
        sample app is already running properly on the device.
    """
    super().__init__(
        auto_reopen=auto_reopen,
        open_on_start=open_on_start)
    self.comms_address = comms_address
    self._protobuf_import_paths = protobuf_import_paths
    self._address = (comms_address, port)
    self._socket = None
    self._hdlc_client = None

  def is_open(self) -> bool:
    """Returns True if the PwRPC transport is connected to the target.

    Returns:
      True if transport is open, False otherwise.
    """
    return self._hdlc_client is not None and self._hdlc_client.is_alive()

  def _close(self) -> None:
    """Closes the PwRPC transport."""
    self._hdlc_client.close()
    self._socket.close()

  def _open(self) -> None:
    """Opens the PwRPC transport."""
    # We always create new socket instance given that reconnecting to a closed
    # socket is not allowed:
    # https://docs.python.org/2/library/socket.html#socket.socket.close
    self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self._hdlc_client = PwHdlcRpcClient(
        self._socket, self._protobuf_import_paths)
    self._socket.connect(self._address)
    self._hdlc_client.start()

  def _read(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, size: int, timeout: float) -> bytes:
    """Returns Pigweed log from the HDLC channel 1.

    Args:
      size: Not used.
      timeout: Maximum seconds to wait to read bytes.

    Returns:
      bytes read from transport or None if no bytes were read
    """
    # Retrieving logs from queue doesn't support size configuration.
    del size  # not used
    try:
      return self._hdlc_client.log_queue.get(timeout=timeout)
    except queue.Empty:
      return b""

  def _write(self, data: str, timeout: Optional[float] = None) -> int:
    """Dummy method for the abstract parent class."""
    del data  # not used
    del timeout  # not used
    return 0

  def rpc(self,
          service_name: str,
          event_name: str,
          **kwargs: Any) -> bytes:
    """RPC call to the Matter endpoint with given service and event name."""
    return _rpc(self._hdlc_client, service_name, event_name, **kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter linux sample app controls and management capability over shell."""
import os
import time
from typing import Callable, List

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import matter_sample_app_base
from gazoo_device.utility import pwrpc_utils
from gazoo_device.utility import retry
import immutabledict

logger = gdm_logger.get_logger()

_MATTER_APP_PATH = f"/home/ubuntu/{pwrpc_utils.MATTER_LINUX_APP_NAME}"
_MATTER_APP_SERVICE = f"{pwrpc_utils.MATTER_LINUX_APP_NAME}.service"
_MATTER_APP_SERVICE_FULL_PATH = f"/etc/systemd/system/{_MATTER_APP_SERVICE}"
_APP_START_SEC = 3  # seconds
_POLL_INTERVAL_SEC = 1  # seconds
_COOL_DOWN_SEC = 1  # seconds
COMMANDS = immutabledict.immutabledict({
    "CHECK_IF_APP_EXISTS": f"test -f {_MATTER_APP_PATH}",
    "CHECK_IF_SERVICE_EXISTS": f"test -f {_MATTER_APP_SERVICE_FULL_PATH}",
    "CHECK_SERVICE_ENABLED": f"sudo systemctl is-enabled {_MATTER_APP_SERVICE}",
    "GET_RUNNING_PROCESSES": f"pgrep -f {pwrpc_utils.MATTER_LINUX_APP_NAME}",
    "TERMINATE_PROCESS": "sudo kill -9 {}",
    "ENABLE_SERVICE": f"sudo systemctl enable {_MATTER_APP_SERVICE}",
    "START_SERVICE": f"sudo systemctl start {_MATTER_APP_SERVICE}",
    "STOP_SERVICE": f"sudo systemctl stop {_MATTER_APP_SERVICE}",
    "CLEAR_COMMISSION_DIRECTORY": "sudo rm -rf /tmp/chip*",
})


class MatterSampleAppShell(matter_sample_app_base.MatterSampleAppBase):
  """Matter app controls and management capability over shell."""

  def __init__(
      self,
      device_name: str,
      shell_fn: Callable[..., str],
      close_transport_fn: Callable[[int], None],
      open_transport_fn: Callable[[int], None],
      pigweed_port: int,
      send_file_to_device_fn: Callable[..., None],
      wait_for_bootup_complete_fn: Callable[[int], None],
      reset_endpoints_fn: Callable[[], None]):
    """Initializes an instance of MatterSampleAppShell capability.

    Args:
      device_name: Name of device using this capability.
      shell_fn: The shell method of the device.
      close_transport_fn: switchboard.close_transport method.
      open_transport_fn: switchboard.open_transport method.
      pigweed_port: Pigweed RPC port number.
      send_file_to_device_fn: The send_file_to_device method.
      wait_for_bootup_complete_fn: The wait_for_bootup_complete method.
      reset_endpoints_fn: The matter_endpoinst.reset method.
    """
    super().__init__(device_name=device_name)
    self._shell = shell_fn
    self._close_transport = close_transport_fn
    self._open_transport = open_transport_fn
    self._pigweed_port = pigweed_port
    self._send_file_to_device = send_file_to_device_fn
    self._wait_for_bootup_complete = wait_for_bootup_complete_fn
    self._reset_endpoints = reset_endpoints_fn

  @decorators.DynamicProperty
  def is_present(self) -> bool:
    """Returns if the app binary presents."""
    _, return_code = self._shell(
        COMMANDS["CHECK_IF_APP_EXISTS"], include_return_code=True)
    return return_code == 0

  @decorators.DynamicProperty
  def is_running(self) -> bool:
    """Returns if the app process is running."""
    return bool(self.get_process_ids())

  @decorators.DynamicProperty
  def has_service(self) -> bool:
    """Returns if the sample app service file exists."""
    _, return_code = self._shell(
        COMMANDS["CHECK_IF_SERVICE_EXISTS"], include_return_code=True)
    return return_code == 0

  @decorators.DynamicProperty
  def is_service_enabled(self) -> bool:
    """Returns if the sample app service is enabled."""
    return self._shell(COMMANDS["CHECK_SERVICE_ENABLED"]) == "enabled"

  @decorators.CapabilityLogDecorator(logger)
  def get_process_ids(self) -> List[str]:
    """Gets all currently active linux app process IDs.

    Returns:
      The list of active linux app process IDs.
    """
    output = self._shell(COMMANDS["GET_RUNNING_PROCESSES"])
    return list(output.splitlines()) if output else []

  @decorators.CapabilityLogDecorator(logger)
  def restart(self) -> None:
    """Restarts the linux app process.

    Note that the Matter sample app process is supposed to run as a linux system
    service, so no need to restart it manually after sending a termination
    command.

    Raises:
      CommunicationTimeoutError when the sample app didn't restart successfully.
    """
    self._close_transport(port=self._pigweed_port)

    # Terminate the process and wait for it to restart.
    running_processes = " ".join(self.get_process_ids())
    terminate_proc_cmd = (
        COMMANDS["TERMINATE_PROCESS"].format(running_processes))
    self._shell(terminate_proc_cmd)

    # Poll until the sample app process is up.
    retry.retry(
        func=self.get_process_ids,
        timeout=_APP_START_SEC,
        interval=_POLL_INTERVAL_SEC,
        is_successful=bool,
        reraise=False)

    self._open_transport(port=self._pigweed_port)

    # Wait for the sample app to become responsive to RPCs.
    self._wait_for_bootup_complete()

  @decorators.CapabilityLogDecorator(logger)
  def enable_service(self) -> None:
    """Enables the sample app service.

    Does nothing if the service is already enabled.
    """
    if not self.is_service_enabled:
      self._shell(COMMANDS["ENABLE_SERVICE"])

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self, sample_app_file_path: str) -> None:
    """Upgrades to a new sample app binary.

    Args:
      sample_app_file_path: Local path to the sample app file on the host.

    Raises:
      FileNotFoundError: when the given sample app file does not exist.
      PermissionError: when the sample app file is not executable.
    """
    if not os.path.exists(sample_app_file_path):
      raise FileNotFoundError(
          f"Sample app file {sample_app_file_path} does not exist.")

    if not os.access(sample_app_file_path, os.X_OK):
      raise PermissionError(
          f"Sample app file {sample_app_file_path} is not executable.")

    self._close_transport(port=self._pigweed_port)

    # Stop the service daemon, scp the sample app binary and start again.
    self._shell(COMMANDS["STOP_SERVICE"])
    self._send_file_to_device(src=sample_app_file_path, dest=_MATTER_APP_PATH)
    self._shell(COMMANDS["START_SERVICE"])

    # Poll until the sample app process is up.
    retry.retry(
        func=self.get_process_ids,
        timeout=_APP_START_SEC,
        interval=_POLL_INTERVAL_SEC,
        is_successful=bool,
        reraise=False)

    # Sometimes process being up doesn't mean it's able to connect.
    # A cool down period is needed to prevent connection failure.
    time.sleep(_COOL_DOWN_SEC)

    self._open_transport(port=self._pigweed_port)

    # Wait for the sample app to become responsive to RPCs and reset the cached
    # endpoints.
    self._wait_for_bootup_complete()
    self._reset_endpoints()

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self) -> None:
    """Factory resets all settings stored on RPi."""
    self._shell(COMMANDS["CLEAR_COMMISSION_DIRECTORY"])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Matter Root Node endpoint.

This endpoint module corresponds to the
"Root Node" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Root Node endpoint houses an instance of Root Node Matter device type with
the following defined clusters on this endpoint:

The required clusters for this endpoint: Basic Information, Access Control,
  Group Key Management, General Commissioning, Administrator Commissioning,
  Node Operational Credentials, General Diagnostics.
The optional clusters for this endpoint: Power Source Configuration, Time
  Synchronization, Diagnostic Logs, Software Diagnostics.
"""

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import basic_information_chip_tool
from gazoo_device.capabilities.matter_endpoints.interfaces import root_node_base


class RootNodeEndpoint(root_node_base.RootNodeBase):
  """Matter Root Node endpoint."""

  @decorators.CapabilityDecorator(
      basic_information_chip_tool.BasicInformationClusterChipTool)
  def basic_information(
      self) -> basic_information_chip_tool.BasicInformationClusterChipTool:
    """Matter Root Node cluster instance."""
    return self.cluster_lazy_init(matter_enums.BasicInformationCluster.ID)

  # TODO(b/241164443): implement additional clusters for root node.

  # def access_control(self):
  #   """Matter access control cluster instance."""

  # def group_key_management(self):
  #   """Matter group key management cluster instance."""

  # def general_commissioning(self):
  #   """Matter general commissioning cluster instance."""

  # def administrator_commissioning(self):
  #   """Matter administrator commissioning cluster instance."""

  # def node_operational_credentials(self):
  #   """Matter node_operational_credentials cluster instance."""

  # def general_diagnostics(self):
  #   """Matter general diagnostics cluster instance."""

  # def power_source_configuration(self):
  #   """Matter power source configuration cluster instance."""

  # def time_synchronization(self):
  #   """Matter time synchronization cluster instance."""

  # def diagnostic_logs(self):
  #   """Matter diagnostic logs cluster instance."""

  # def software_diagnostics(self):
  #   """Matter software diagnostics cluster instance."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility module for HTTP GET and POST methods."""
import contextlib
import json
import socket
import ssl
from typing import Any, Callable, ContextManager, Dict, List, Optional, Union
import urllib

from gazoo_device import gdm_logger
import requests
from requests import adapters
from requests.auth import AuthBase
import urllib3
from urllib3 import connection
from urllib3 import connectionpool
from urllib3 import poolmanager

MAX_URL_RETRY = 3

logger = gdm_logger.get_logger()


class SSLAdapter(adapters.HTTPAdapter):
  """An HTTPS Transport Adapter that uses an arbitrary SSL version."""

  def __init__(self, ssl_version: Optional[int] = None, **kwargs: Any):
    self.ssl_version = ssl_version

    super(SSLAdapter, self).__init__(**kwargs)

  def init_poolmanager(self,
                       connections: int,
                       maxsize: int,
                       block: bool = False) -> None:
    self.poolmanager = poolmanager.PoolManager(
        num_pools=connections,
        maxsize=maxsize,
        block=block,
        ssl_version=self.ssl_version)


def send_http_get(url: str,
                  auth: Optional[AuthBase] = None,
                  params: Optional[Dict[Any, Any]] = None,
                  data: Optional[Any] = None,
                  headers: Optional[Dict[Any, Any]] = None,
                  ssl_version: Optional[int] = None,
                  valid_return_codes: Optional[List[int]] = None,
                  session_factory: Optional[Callable[
                      [], ContextManager[requests.Session]]] = requests.Session,
                  timeout: int = 10,
                  tries: int = 1) -> requests.Response:
  """Issues a HTTP GET and returns the response if the request is successful.

  Args:
      url: HTTP URL to which HTTP GET request needs to be sent
      auth: HTTP authentication object (i.e. HTTPDigestAuth)
      params: Parameters to send as a query string.
      data: Data for this HTTP GET Request
      headers: Headers for this HTTP GET Request
      ssl_version: SSL version to be used for secure http (https) requests. For
        example, ssl.PROTOCOL_TLSv1_2.
      valid_return_codes: List of valid HTTP return codes.
      session_factory: A context manager to yield a session to be used.
      timeout: request timeout in seconds
      tries: how many times to try sending the request.

  Raises:
      RuntimeError: if response.status_code returned by requests.get() is not in
          valid_return_codes
      TypeError: if headers is not a dictionary
      RequestException, HTTPError: raised if request fails and are converted
          into a RuntimeError.

  Returns:
      requests.Response: Response returned by requests.get
  """
  valid_return_codes = valid_return_codes or [200]

  if headers and not isinstance(headers, dict):
    raise TypeError(
        "Expecting a dict value in headers param but received: {}".format(
            type(headers)))
  try:
    with session_factory() as session:
      session: requests.Session
      if ssl_version:
        session.mount("https://", SSLAdapter(ssl_version))
      if data and isinstance(data, dict):
        data = json.dumps(data)

      for attempt in range(tries):
        try:
          response: requests.Response = session.get(
              url,
              auth=auth,
              params=params,
              data=data,
              headers=headers,
              timeout=timeout,
              verify=False)
          break
        except (requests.exceptions.RequestException,
                urllib3.exceptions.HTTPError) as err:
          if attempt < tries - 1:
            logger.debug(
                "HTTP GET to URL {} with data {} failed. Retrying. Error: {!r}"
                .format(url, data, err))
          else:
            raise

  except Exception as err:
    raise RuntimeError("HTTP GET to URL {} with data {} failed".format(
        url, data)) from err

  if response.status_code not in valid_return_codes:
    raise RuntimeError(
        f"HTTP GET to URL {url} "
        f"status code: {response.status_code}, reason: {response.reason}"
    )

  return response


def send_http_post(url: str,
                   auth: Optional[AuthBase] = None,
                   data: Optional[Any] = None,
                   params: Optional[Dict[Any, Any]] = None,
                   headers: Optional[Dict[Any, Any]] = None,
                   json_data: Optional[Union[Dict[Any, Any], List[Any]]] = None,
                   ssl_version: Optional[int] = None,
                   valid_return_codes: Optional[List[int]] = None,
                   session_factory: Optional[Callable[
                       [],
                       ContextManager[requests.Session]]] = requests.Session,
                   timeout: int = 10,
                   tries: int = 1) -> requests.Response:
  """Issues a HTTP POST and returns the response if the request is successful.

  Args:
      url: HTTP URL to which HTTP POST request needs to be sent
      auth: HTTP authentication object (i.e. HTTPDigestAuth)
      data: Data for this HTTP POST Request
      params: Parameters for this HTTP POST Request
      headers: Headers for this HTTP POST Request
      json_data: JSON data for this HTTP POST Request
      ssl_version: SSL version to be used for secure http (https) requests. For
        example, ssl.PROTOCOL_TLSv1_2
      valid_return_codes: List of valid HTTP return codes.
      session_factory: A context manager to yield a session to be used.
      timeout: request timeout in seconds
      tries: how many times to try sending the request.

  Raises:
      RuntimeError: if response.status_code returned by requests.post() is not
          in valid_return_codes
      TypeError: if headers / json_data is not a dictionary or None.
      RequestException, HTTPError: raised if request fails and are converted
          into a RuntimeError.

  Returns:
      requests.Response: Response returned by requests.post
  """
  valid_return_codes = valid_return_codes or [200]

  if json_data is not None and not isinstance(json_data, (dict, list)):
    raise TypeError(
        "Expecting a dict or list value in json_data params but received: "
        "{}.".format(type(json_data)))

  if headers is not None and not isinstance(headers, dict):
    raise TypeError("Expecting a dict value in headers params but received: "
                    "{}.".format(type(headers)))

  try:
    with session_factory() as session:
      session: requests.Session
      if ssl_version:
        session.mount("https://", SSLAdapter(ssl_version))

      for attempt in range(tries):
        try:
          response: requests.Response = session.post(
              url,
              auth=auth,
              data=data,
              json=json_data,
              headers=headers,
              timeout=timeout,
              params=params,
              verify=False)
          break
        except (requests.exceptions.RequestException,
                urllib3.exceptions.HTTPError) as err:
          if attempt < tries - 1:
            logger.debug(
                "HTTP POST to URL {} with headers {}, data {} and json data {} "
                "failed. Retrying. Error: {!r}".format(url, headers, data,
                                                       json_data, err))
          else:
            raise

  except Exception as err:
    raise RuntimeError(
        "HTTP POST to URL {} with headers {}, data {} and json data {} failed"
        .format(url, headers, data, json_data)) from err

  if response.status_code not in valid_return_codes:
    raise RuntimeError(
        f"HTTP POST to URL {url} with headers {headers}, data {data} and json"
        f" data {json_data} status code: {response.status_code}, reason:"
        f" {response.reason}"
    )

  return response


def is_valid_ip_address(address: str, is_ipv6: bool = False) -> bool:
  """Checks if valid IP address.

  Args:
      address: IP address
      is_ipv6: if IP address is IPv6

  Returns:
      bool: True if valid IP else False
  """
  address_family = socket.AF_INET6 if is_ipv6 else socket.AF_INET
  if is_ipv6 and "%" in address:  # Check for link local Ipv6 address
    # Example: fe80::123:1234:1234:123%en0
    address = address.split("%")[0]
  try:
    socket.inet_pton(address_family, address)
  except socket.error:  # not a valid address
    return False
  return True


def validate_url_access(url: str) -> None:
  """Validates URL access.

  Args:
      url: URL to access.

  Raises:
      RuntimeError: if unable to access domain of URL.
  """
  parsed_uri = urllib.parse.urlparse(url)
  prefix = "{uri.scheme}://{uri.netloc}/".format(uri=parsed_uri)
  response_code = urllib.request.urlopen(prefix).getcode()
  if response_code != 200:
    raise RuntimeError("Unable to reach domain {}. Returned code {}".format(
        prefix, response_code))


class UnixSocketHTTPSConnection(connection.HTTPSConnection):
  """A HTTPS connection via a unix socket."""

  def __init__(self, path: str):
    # Device servers are using self-signed certificates.
    # Do not verify their certificates.
    self.cert_reqs = ssl.CERT_NONE
    self.path = path
    super().__init__("localhost")

  def _new_conn(self):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(self.path)
    return sock


class UnixSocketHTTPSConnectionPool(connectionpool.HTTPSConnectionPool):

  def __init__(self, path: str):
    self.path = path
    super().__init__("localhost")

  def _new_conn(self):
    return UnixSocketHTTPSConnection(self.path)


class UnixSocketHTTPSAdapter(adapters.HTTPAdapter):

  def __init__(self, path: str):
    self.path = path
    super().__init__()

  def get_connection(self, url, proxies=None):
    return UnixSocketHTTPSConnectionPool(self.path)


class UnixSocketHTTPConnection(connection.HTTPConnection):
  """A HTTP connection via a unix socket."""

  def __init__(self, path: str):
    self.path = path
    super().__init__("localhost")

  def _new_conn(self):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(self.path)
    return sock


class UnixSocketHTTPConnectionPool(connectionpool.HTTPConnectionPool):

  def __init__(self, path: str):
    self.path = path
    super().__init__("localhost")

  def _new_conn(self):
    return UnixSocketHTTPConnection(self.path)


class UnixSocketHTTPAdapter(adapters.HTTPAdapter):

  def __init__(self, path: str):
    self.path = path
    super().__init__()

  def get_connection(self, url, proxies=None):
    return UnixSocketHTTPConnectionPool(self.path)


@contextlib.contextmanager
def unix_socket_http_session(path: str):
  """Initializes an HTTP or HTTPS session via a unix socket.

  Args:
    path: The unix socket to handle HTTP(S) requests.

  Yields:
    The requests session.
  """

  with requests.Session() as session:
    session.mount("udss://localhost/", UnixSocketHTTPSAdapter(path))
    session.mount("uds://localhost/", UnixSocketHTTPAdapter(path))
    yield session



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the JLink flashing capability."""
import os
from typing import Callable, List, NoReturn, Optional, Tuple

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import flash_build_base
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.utility import subprocess_utils
import immutabledict
import intelhex
import pylink

logger = gdm_logger.get_logger()
UNKNOWN = flash_build_base.UNKNOWN
_FLASH_TIMEOUT_S = 180
_JLINK_NO_DLL_ERROR = "Expected to be given a valid DLL."


class FlashBuildJLink(flash_build_base.FlashBuildBase):
  """J-Link implementation of flash_build capability."""

  def __init__(
      self,
      device_name: str,
      serial_number: str,
      platform_name: str,
      reset_endpoints_fn: Optional[Callable[[], None]] = None,
      switchboard: Optional[switchboard_base.SwitchboardBase] = None,
      wait_for_bootup_complete_fn: Optional[Callable[[], None]] = None):
    """Initializes an instance of the FlashBuildJLink capability.

    Args:
      device_name: Device name used for logging.
      serial_number: Device serial number.
      platform_name: The target device's platform name.
      reset_endpoints_fn: Method to reset matter_endpoint capability.
        This method will be called after flashing is completed.
      switchboard: A Switchboard capability instance if the device supports it.
      wait_for_bootup_complete_fn: wait_for_bootup_complete method for verifying
        device responsiveness after flashing.
    """
    super().__init__(device_name=device_name)
    self._serial_number = serial_number
    self._platform_name = platform_name
    self._reset_endpoints_fn = reset_endpoints_fn
    self._switchboard = switchboard
    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._flash_method_map = immutabledict.immutabledict({
        False: self._flash_device,
    })

  @decorators.CapabilityLogDecorator(logger)
  def flash_device(self,
                   list_of_files: List[str],
                   expected_version: Optional[str] = None,
                   expected_build_type: Optional[str] = None,
                   verify_flash: bool = True,
                   method: Optional[str] = None) -> None:
    """Flashes the firmware image (.hex file) on the device.

    Args:
      list_of_files: Image files on local host, currently supports flashing
        only one hex file at a time.
      expected_version: Not used.
      expected_build_type: Not used.
      verify_flash: Check if we should verify build after flashing.
      method: Not used.

    Raises:
      ValueError: If invalid arguments are provided.
      DeviceError: If flashing fails.
      DependencyUnavailableError: If J-Link SDK is not installed.
    """
    del expected_version, expected_build_type, method  # Unused.
    if len(list_of_files) != 1:
      raise ValueError("Only one hex file can be flashed via JLink.")
    image_path = list_of_files[0]
    if not image_path.endswith(".hex"):
      raise ValueError("Only hex type file can be flashed.")
    if not os.path.exists(image_path):
      raise ValueError(f"Firmware image {image_path} does not exist.")

    flash_method_key = False
    flash_method = self._flash_method_map[flash_method_key]
    flash_method(image_path, verify_flash)

  def _flash_device(self, image_path: str, verify_flash: bool) -> None:
    """Flashes device via pylink module."""
    try:
      jlink = pylink.JLink()
    except TypeError as e:
      if _JLINK_NO_DLL_ERROR in str(e):
        raise errors.DependencyUnavailableError(
            "No J-Link DLL found. Install the J-Link SDK from "
            "https://www.segger.com/downloads/jlink/#J-LinkSoftwareAndDocumentationPack. "
            f"Error: {e!r}.")
      else:
        raise

    self._pre_flashing(jlink)
    self._flash_image(jlink, image_path)
    self._post_flashing(jlink)

    # For Matter device classes, we'll need to reset the Matter endpoint mapping
    # as the supported endpoints might change after flashing a new build.
    if self._reset_endpoints_fn is not None:
      self._reset_endpoints_fn()
    if verify_flash and self._wait_for_bootup_complete_fn is not None:
      self._wait_for_bootup_complete_fn()

  def _pre_flashing(self, jlink: pylink.jlink.JLink) -> None:
    """Opens Jlink connection and connects to the board."""
    jlink.open(serial_no=self._serial_number)
    jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)
    jlink.connect(chip_name=self._platform_name, speed="auto")
    jlink.halt()

  def _post_flashing(self, jlink: pylink.jlink.JLink) -> None:
    """Resets and closes the Jlink connection."""
    jlink.reset()
    jlink.restart()
    jlink.close()

  def _flash_image(self, jlink: pylink.jlink.JLink, image_path: str) -> None:
    """Flashes image onto the board."""
    image = intelhex.IntelHex(os.path.abspath(image_path))
    # Find the first segment
    segment_start, _ = image.segments()[0]
    jlink.flash_file(image_path, segment_start)

  @decorators.CapabilityLogDecorator(logger)
  def extract_build_info(self, *args, **kwargs) -> NoReturn:
    """Converts the provided build arguments into info about the build."""
    raise NotImplementedError(
        "extract_build_info is not available in flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self, build_file: str) -> None:
    """Upgrade the device based on the provided build arguments.

    Args:
      build_file: Local path to the file.
    """
    self.flash_device(list_of_files=[build_file])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Build Flasher capability interface."""

import abc
import dataclasses
from typing import Any, Optional, Sequence, Tuple

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import capability_base

_LOGGER = gdm_logger.get_logger()
UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True)
class BuildInfo:
  build_version: str
  build_type: str
  build_branch: str
  build_files: Tuple[str, ...]
  is_local: bool


class FlashBuildBase(capability_base.CapabilityBase):
  """Abstract base class for the build_flasher capability."""

  @abc.abstractmethod
  def upgrade(
      self,
      *unexpected_args,
      build_folder_url: Optional[str] = None,
      local_files: Optional[Sequence[str]] = None,
      forced_upgrade: bool = False,
      method: Optional[str] = None,
      file_regexes: Optional[Sequence[str]] = None,
      num_files: Optional[int] = None,
      **unexpected_kwargs: Any,
  ) -> None:
    """Upgrade the device based on the provided build arguments.

    Args:
      *unexpected_args: Captures unexpected positional arguments.
      build_folder_url: gs:// URL of the folder the build files reside in.
      local_files: Full paths to local build files.
        Mutually exclusive with build_folder_url and file_regexes.
      forced_upgrade: Reflash the firmware even if the same build is already
        installed on the device.
      method: Used by some flavors to pick between flashing methods.
      file_regexes: Regexes for finding the build files in a folder. If not
        provided, uses the default in nest_device/build_defaults/info.json
      num_files: The number of build files expected after regex matching.
      **unexpected_kwargs: Captures unexpected keyword arguments.

    Raises:
      DeviceError: Both upgrades over_the_wire & over_the_air are unsupported.
    """

  @abc.abstractmethod
  def extract_build_info(
      self,
      *unexpected_args,
      build_folder_url: Optional[str] = None,
      local_files: Optional[Sequence[str]] = None,
      forced_upgrade: bool = False,
      method: Optional[str] = None,
      file_regexes: Optional[Sequence[str]] = None,
      num_files: Optional[int] = None,
      **unexpected_kwargs: Any,
  ) -> BuildInfo:
    """Converts the provided build arguments into information about the build.

    Args:
      *unexpected_args: Captures unexpected positional arguments.
      build_folder_url: gs:// URL of the folder the remote build files reside
        in.
      local_files: Full paths to local build files.
        Mutually exclusive with build_folder_url and file_regexes.
      forced_upgrade: No-op. Not used by extract_build_info. The arg is here
        so that the method signature is identical to upgrade().
      method: Used by some flavors to pick between flashing methods.
      file_regexes: Regexes for finding the build files in a folder. If not
        provided, uses the default in nest_device/build_defaults/info.json
      num_files: The number of build files expected after regex matching.
      **unexpected_kwargs: Captures unexpected keyword arguments.

    Returns:
      BuildInfo about the requested build's version, type, and file locations.
      For upgrades from build folder URLs, this method transforms
      "build_folder_url" + "file_regexes" into remote build file locations.
    """

  @decorators.CapabilityLogDecorator(_LOGGER)
  def flash_device(
      self,
      list_of_files: Sequence[str],
      expected_version: Optional[str] = None,
      expected_build_type: Optional[str] = None,
      verify_flash: bool = True,
      method: Optional[str] = None,
  ) -> None:
    """Flashes the device over the wire given paths to build files on the host.

    This method will not delete the files provided in list_of_files.

    Args:
      list_of_files: List of local build files.
      expected_version: Expected firmware_version for the build.
      expected_build_type: Expected firmware_type for the build.
      verify_flash: If True, does a post-upgrade verification check of the
        expected vs actual build flashed on the device.
      method: Upgrade method.

    Raises:
      NotImplementedError: must be overridden in derived classes which support
        flashing over the wire.
    """
    raise NotImplementedError(f"flash_device not implemented in {type(self)}")

  @decorators.CapabilityLogDecorator(_LOGGER)
  def ota_device(
      self,
      list_of_files: Sequence[str],
      expected_version: Optional[str] = None,
      expected_build_type: Optional[str] = None,
      verify_flash: bool = True,
      method: Optional[str] = None,
  ) -> None:
    """Flashes the device over the air given paths to build files on the host.

    This method will not delete the files provided in list_of_files.

    Args:
      list_of_files: List of local build files.
      expected_version: Expected firmware_version for the build.
      expected_build_type: Expected firmware_type for the build.
      verify_flash: If True, does a post-upgrade verification check of the
        expected vs actual build flashed on the device.
      method: Upgrade method.

    Raises:
      NotImplementedError: must be overridden in derived classes which support
        flashing over the air.
    """
    raise NotImplementedError(f"ota_device not implemented in {type(self)}")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""D-Link switch device controller."""
import re
import subprocess
from typing import Any, Dict, Tuple

from gazoo_device import custom_types
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import switch_power_snmp
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()

_GET_DLINK_TOTAL_PORTS_SNMP_COMMAND = (
    "snmpget -v 2c -c private {ip_address}:161 1.3.6.1.2.1.2.1.0")
_DLINK_TOTAL_PORTS_RESPONSE_REG_EX = r".+INTEGER:.+?(\d+)"
_SNMP_TIMEOUT_S = 10


class DLinkSwitch(auxiliary_device.AuxiliaryDevice):
  """Device class for a D-Link Switch."""
  COMMUNICATION_TYPE = "SnmpComms"
  DETECT_MATCH_CRITERIA = {detect_criteria.SnmpQuery.IS_DLINK: True}
  DEVICE_TYPE = "dlink_switch"
  _OWNER_EMAIL = "gdm-authors@google.com"

  def _get_total_ports(self) -> int:
    """Gets the total number of network ports.

    Returns:
      Total number of network ports.

    Raises:
      DeviceError: When the response to querying the SNMP total ports OID
        is unexpected.
      CalledProcessError: When the snmp shell command sent to the dlink's
        given ip_address is unsuccessful.
    """
    command = _GET_DLINK_TOTAL_PORTS_SNMP_COMMAND.format(
        ip_address=self.ip_address)
    # Expected response for supported dlink switch total ports should look like:
    # "iso.3.6.1.2.1.2.1.0 = INTEGER: <total-number-of-ports>".
    response = subprocess.check_output(
        command.split(), text=True, timeout=_SNMP_TIMEOUT_S)
    logger.debug(
        "%s sent command %s and got response %r", self.name, command, response)
    total_ports = re.search(_DLINK_TOTAL_PORTS_RESPONSE_REG_EX, response)
    if total_ports:
      return int(total_ports[1])
    raise errors.DeviceError(f"Failed to retrieve total ports from "
                             f"dlink_switch {self.name} with command: "
                             f"{command}. Unexpected response: {response}")

  @decorators.PersistentProperty
  def ip_address(self):
    """Global IP address."""
    return self.communication_address

  @decorators.LogDecorator(logger)
  def get_console_configuration(self) -> None:
    """Returns None. Console is not supported because there's no Switchboard."""
    del self  # Unused because console is not supported.
    return None

  @decorators.CapabilityDecorator(switch_power_snmp.SwitchPowerSnmp)
  def switch_power(self):
    """Capability to control switch power setting over SNMP."""
    return self.lazy_init(
        switch_power_snmp.SwitchPowerSnmp,
        device_name=self.name,
        ip_address=self.ip_address,
        total_ports=self.total_ports)

  @decorators.LogDecorator(logger)
  def get_detection_info(self) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Gets the persistent and optional attributes of a DLI Power Switch.

    Returns:
      Tuple of dictionaries containing persistent and optional attributes.
    """
    self.props["optional"] = {}
    self.props["persistent_identifiers"]["model"] = (
        detect_criteria.get_dlink_model_name(self.ip_address))
    # Cannot get serial number via snmp; no other known way for this device.
    # Using IP as a unique identifier instead.
    self.props["persistent_identifiers"]["serial_number"] = (
        self.ip_address.replace(".", ""))
    self.props["persistent_identifiers"]["total_ports"] = (
        self._get_total_ports())
    return self.props["persistent_identifiers"], self.props["optional"]

  @classmethod
  def is_connected(cls,
                   device_config: custom_types.ManagerDeviceConfigDict) -> bool:
    """Checks whether or not the D-Link Power Switch is connected.

    Args:
      device_config: Contains the persistent attributes.

    Returns:
      True if device responds to a ping. False if not.
    """
    return host_utils.is_pingable(
        device_config["persistent"]["console_port_name"])

  @decorators.PersistentProperty
  def total_ports(self) -> int:
    """Returns the dlink switch's total number of ports."""
    return self.props["persistent_identifiers"]["total_ports"]



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""chip-tool implementation of Matter Basic Information cluster capability.
"""

from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.matter_clusters.interfaces import basic_information_base

logger = gdm_logger.get_logger()
_CLUSTER_NAME = "basic"


class BasicInformationClusterChipTool(
    basic_information_base.BasicInformationClusterBase):
  """Matter Basic Information cluster capability."""

  @decorators.DynamicProperty
  def data_model_revision(self) -> int:
    """The DataModelRevision attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "data-model-revision")

  @decorators.DynamicProperty
  def vendor_name(self) -> str:
    """The VendorName attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "vendor-name")

  @decorators.DynamicProperty
  def vendor_id(self) -> int:
    """The VendorID attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "vendor-id")

  @decorators.DynamicProperty
  def product_name(self) -> str:
    """The ProductName attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "product-name")

  @decorators.DynamicProperty
  def product_id(self) -> int:
    """The ProductID attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "product-id")

  @decorators.DynamicProperty
  def node_label(self) -> str:
    """The NodeLabel attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "node-label")

  @decorators.DynamicProperty
  def location(self) -> str:
    """The Location attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "location")

  @decorators.DynamicProperty
  def hardware_version(self) -> int:
    """The HardwareVersion attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "hardware-version")

  @decorators.DynamicProperty
  def hardware_version_string(self) -> str:
    """The HardwareVersionString attribute."""
    return self._read(
        self._endpoint_id, _CLUSTER_NAME, "hardware-version-string")

  @decorators.DynamicProperty
  def software_version(self) -> int:
    """The SoftwareVersion attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, "software-version")

  @decorators.DynamicProperty
  def software_version_string(self) -> str:
    """The SoftwareVersionString attribute."""
    return self._read(
        self._endpoint_id, _CLUSTER_NAME, "software-version-string")

  @decorators.DynamicProperty
  def capability_minima(self) -> int:
    """The CapabilityMinima attribute."""
    return self._read(
        self._endpoint_id, _CLUSTER_NAME, "capability-minima")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Fan endpoint.

This endpoint module corresponds to the
"Fan" Matter device type which defines required and optional clusters
in the Matter spec 1.0.

A Fan enpoint houses an instance of Fan Matter device type with
the following defined clusters on this endpoint:

The required clusters for this endpoint: Identify, Groups and FanControl
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import fan_control_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import fan_base


class FanEndpoint(fan_base.FanBase):
  """Matter Fan endpoint."""

  @decorators.CapabilityDecorator(fan_control_pw_rpc.FanControlClusterPwRpc)
  def fan_control(self) -> fan_control_pw_rpc.FanControlClusterPwRpc:
    """Matter Fan Control cluster instance."""
    return self.cluster_lazy_init(matter_enums.FanControlCluster.ID)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""chip-tool implementation of the Matter Thermostat cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import thermostat_base

logger = gdm_logger.get_logger()

_CLUSTER_NAME = "thermostat"


class ThermostatClusterChipTool(thermostat_base.ThermostatClusterBase):
  """Matter Thermostat cluster capability."""

  @decorators.DynamicProperty
  def local_temperature(self) -> int:
    """The LocalTemperature attribute.

    Returns:
      The temperature, as measured locally or remotely (over the network),
      on the Celsius scale with a resolution of 0.01°C.
    """
    return self._read(self._endpoint_id, _CLUSTER_NAME, "local-temperature")

  @decorators.DynamicProperty
  def occupied_cooling_setpoint(self) -> int:
    """The OccupiedCoolingSetpoint attribute.

    Returns:
      The cooling mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """
    return self._read(self._endpoint_id, _CLUSTER_NAME,
                      "occupied-cooling-setpoint")

  @occupied_cooling_setpoint.setter
  def occupied_cooling_setpoint(self, value: int) -> None:
    """Updates the OccupiedCoolingSetpoint attribute with new value."""
    self._write(self._endpoint_id, _CLUSTER_NAME, "occupied-cooling-setpoint",
                value)

  @decorators.DynamicProperty
  def occupied_heating_setpoint(self) -> int:
    """The OccupiedHeatingSetpoint attribute.

    Returns:
      The heating mode setpoint when the room is occupied, on the Celsius scale
      with a resolution of 0.01°C.
    """
    return self._read(self._endpoint_id, _CLUSTER_NAME,
                      "occupied-heating-setpoint")

  @occupied_heating_setpoint.setter
  def occupied_heating_setpoint(self, value: int) -> None:
    """Updates the OccupiedHeatingSetpoint attribute with new value."""
    self._write(self._endpoint_id, _CLUSTER_NAME, "occupied-heating-setpoint",
                value)

  @decorators.DynamicProperty
  def control_sequence_of_operation(
      self) -> matter_enums.ThermostatControlSequence:
    """The ControlSequenceOfOperation attribute.

    Returns:
      The overall operating environment of the thermostat, and the possible
      system modes that the thermostat can operate in.
    """
    operation_data = self._read(self._endpoint_id, _CLUSTER_NAME,
                                "control-sequence-of-operation")
    return matter_enums.ThermostatControlSequence(operation_data)

  @control_sequence_of_operation.setter
  def control_sequence_of_operation(
      self, value: matter_enums.ThermostatControlSequence) -> None:
    """Updates the ControlSequenceOfOperation attribute with new value."""
    self._write(self._endpoint_id, _CLUSTER_NAME,
                "control-sequence-of-operation", value.value)

  @decorators.DynamicProperty
  def system_mode(self) -> matter_enums.ThermostatSystemMode:
    """The SystemMode attribute.

    Returns:
      The current operating mode of the thermostat.
    """
    mode_data = self._read(self._endpoint_id, _CLUSTER_NAME, "system-mode")
    return matter_enums.ThermostatSystemMode(mode_data)

  @system_mode.setter
  def system_mode(self, value: matter_enums.ThermostatSystemMode) -> None:
    """Updates the SystemMode attribute with new value."""
    self._write(self._endpoint_id, _CLUSTER_NAME, "system-mode", value.value)

  @decorators.CapabilityLogDecorator(logger)
  def setpoint_raise_lower(self, mode: matter_enums.ThermostatSetpointMode,
                           amount: int) -> None:
    """The SetpointRaiseLower command to add specified amount to setpoint.

    Args:
      mode: The mode for which the sepoints are to be adjusted.
      amount: The amount (possibly negative) that should be added to the
        setpoint(s), in steps of 0.1°C
    """
    self._send(self._endpoint_id, _CLUSTER_NAME, "setpoint-raise-lower",
               [mode.value, amount])



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pigweed RPC implementation of Matter Illuminance Measurement cluster capability.
"""
from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters.interfaces import measurement_base
from gazoo_device.protos import attributes_service_pb2


class IlluminanceMeasurementClusterPwRpc(measurement_base.MeasurementClusterBase
                                        ):
  """Matter Illuminance Measurement cluster capability."""

  CLUSTER_ID = matter_enums.IlluminanceMeasurementCluster.ID
  MATTER_CLUSTER = matter_enums.IlluminanceMeasurementCluster
  ATTRIBUTE_TYPE = (
      attributes_service_pb2.AttributeType.ZCL_INT16U_ATTRIBUTE_TYPE)

  @decorators.DynamicProperty
  def light_sensor_type(self) -> matter_enums.LightSensorType:
    """Fetches the LightSensorType attribute."""
    data = self._read(
        endpoint_id=self._endpoint_id,
        cluster_id=self.MATTER_CLUSTER.ID,
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_LIGHT_SENSOR_TYPE,
        attribute_type=(
            attributes_service_pb2.AttributeType.ZCL_ENUM8_ATTRIBUTE_TYPE,))
    return matter_enums.LightSensorType(data.data_uint8)

  @light_sensor_type.setter
  def light_sensor_type(self, value: matter_enums.LightSensorType) -> None:
    """Updates the LightSensorType attribute with new value."""
    sensor_type = matter_enums.LightSensorType(value)
    self._write(
        endpoint_id=self._endpoint_id,
        cluster_id=self.MATTER_CLUSTER.ID,
        attribute_id=self.MATTER_CLUSTER.ATTRIBUTE_LIGHT_SENSOR_TYPE,
        attribute_type=(
            attributes_service_pb2.AttributeType.ZCL_ENUM8_ATTRIBUTE_TYPE),
        data_uint8=sensor_type)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for bluetooth service controls capability."""
import abc

from gazoo_device.capabilities.interfaces import capability_base


class BluetoothServiceBase(capability_base.CapabilityBase):
  """Abstract base class for bluetooth service controls capability."""

  @abc.abstractmethod
  def status(self) -> bool:
    """Returns if the bluetooth service is active or not."""

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the bluetooth service."""

  @abc.abstractmethod
  def stop(self) -> None:
    """Stops the bluetooth service."""

  @abc.abstractmethod
  def restart(self) -> None:
    """Restarts the bluetooth service."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Matter Occupancy Sensor endpoint.

An Occupancy Sensor is a measurement and sensing device that is capable of
measuring and reporting the occupancy state in a designated area.

A Occupancy Sensor endpoint houses an instance of Occupancy Sensor Matter device
type with the following defined clusters on this endpoint:

The required clusters for this endpoint: Occupancy Sensing
"""
from typing import Union

from gazoo_device import decorators
from gazoo_device.capabilities import matter_enums
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_chip_tool
from gazoo_device.capabilities.matter_clusters import occupancy_sensing_pw_rpc
from gazoo_device.capabilities.matter_endpoints.interfaces import occupancy_sensor_base


class OccupancySensorEndpoint(occupancy_sensor_base.OccupancySensorBase):
  """Matter Occupancy Sensor endpoint."""

  @decorators.CapabilityDecorator([
      occupancy_sensing_chip_tool.OccupancySensingClusterChipTool,
      occupancy_sensing_pw_rpc.OccupancySensingClusterPwRpc
  ])
  def occupancy_sensing(
      self
  ) -> Union[occupancy_sensing_chip_tool.OccupancySensingClusterChipTool,
             occupancy_sensing_pw_rpc.OccupancySensingClusterPwRpc]:
    """Matter Occupancy Sensing cluster instance."""
    return self.cluster_lazy_init(matter_enums.OccupancySensingCluster.ID)



"""Wpan related functionality for devices running wpanctl.

Implemented in most Gazoo embedded linux devices.
"""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class WpanBase(capability_base.CapabilityBase):
  """Class for interacting with wpanctl interface and getting props."""

  @classmethod
  def list_properties_dynamic(cls):
    """Returns a list of dynamic "gdm properties" for this class.

    Each element of the list is a string, with the name of the getter. getattr()
    can be used to read the value. Currently this returns all of the "@property"
    decorated methods. If we add setter properties we will want to revise this.

    Returns:
         list: "attributes" (which can be called by getattr and will return
         a value).
    """
    return [p for p in dir(cls) if isinstance(getattr(cls, p), property)]

  @property
  @abc.abstractmethod
  def ncp_channel(self) -> str:
    """NCP channel value."""

  @property
  @abc.abstractmethod
  def network_name(self) -> str:
    """Wpan network name."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for a Matter Root Node endpoint."""
import abc
from gazoo_device.capabilities.matter_clusters.interfaces import basic_information_base
from gazoo_device.capabilities.matter_endpoints.interfaces import endpoint_base


class RootNodeBase(endpoint_base.EndpointBase, metaclass=abc.ABCMeta):
  """Matter Root Node endpoint interface."""

  DEVICE_TYPE_ID = 0x0016

  @property
  @abc.abstractmethod
  def basic_information(
      self) -> basic_information_base.BasicInformationClusterBase:
    """Required cluster: basic_information cluster."""

  # TODO(b/241164443): implement additional clusters for root node.

  # @property
  # @abc.abstractmethod
  # def access_control(self):
  #   """Required cluster: access_control cluster."""

  # @property
  # @abc.abstractmethod
  # def group_key_management(self):
  #   """Required cluster: group_key_management cluster."""

  # @property
  # @abc.abstractmethod
  # def general_commissioning(self):
  #   """Required cluster: general_commissioning cluster."""

  # @property
  # @abc.abstractmethod
  # def administrator_commissioning(self):
  #   """Required cluster: administrator_commissioning cluster."""

  # @property
  # @abc.abstractmethod
  # def node_operational_credentials(self):
  #   """Required cluster: node_operational_credentials cluster."""

  # @property
  # @abc.abstractmethod
  # def general_diagnostics(self):
  #   """Required cluster: general_diagnostics cluster."""

  # @property
  # @abc.abstractmethod
  # def power_source_configuration(self):
  #   """Optional cluster: power_source_configuration cluster."""

  # @property
  # @abc.abstractmethod
  # def time_synchronization(self):
  #   """Optional cluster: time_synchronization cluster."""

  # @property
  # @abc.abstractmethod
  # def diagnostic_logs(self):
  #   """Optional cluster: diagnostic_logs cluster."""

  # @property
  # @abc.abstractmethod
  # def software_diagnostics(self):
  #   """Optional cluster: software_diagnostics cluster."""



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for Matter controller capability."""

import abc
import enum
from typing import Any, List, Optional, Sequence

from gazoo_device.capabilities.interfaces import capability_base


@enum.unique
class CommissionMethod(enum.Enum):
  BLE_WIFI = enum.auto()
  BLE_THREAD = enum.auto()
  ON_NETWORK = enum.auto()


class MatterControllerBase(capability_base.CapabilityBase):
  """Abstract base class for Matter controller capability."""

  @property
  @abc.abstractmethod
  def version(self) -> str:
    """Matter SDK version of the controller."""

  @abc.abstractmethod
  def commission(self,
                 node_id: int,
                 setup_code: str,
                 long_discriminator: Optional[int] = None,
                 ssid: Optional[str] = None,
                 password: Optional[str] = None,
                 operational_dataset: Optional[str] = None) -> None:
    """Commissions a device into the controller's fabric.

    Commissioning protocol is based on specified arguments:
      - When operational dataset is provided, pairs the device over ble-thread.
      - When Wi-Fi SSID and password are provided, pairs the device over
        ble-wifi.
      - Otherwise, discover the devices on the network and pairs with the first
        one that matches the setup code and long discriminator if one is
        specified.

    Args:
      node_id: Node ID to assign to the node being commissioned.
      setup_code: Set up PIN code of the remote device.
      long_discriminator: Discriminator of the remote device.
      ssid: Wi-Fi SSID either as a string, or in the form hex:XXXXXXXX where the
        bytes of the SSID are encoded as two-digit hex numbers.
      password: Wi-Fi password, either as a string or as hex data.
      operational_dataset: Thread dataset in base-64. This argument is mutually
        exclusive with ssid and password.
    """

  @abc.abstractmethod
  def decommission(self) -> None:
    """Forgets a commissioned device with the given node id."""

  @abc.abstractmethod
  def read(self, endpoint_id: int, cluster: str, attribute: str) -> Any:
    """Reads a cluster's attribute for the given node id and endpoint.

    Args:
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to read the attribute value from.
      attribute: Name of the cluster attribute to read.

    Returns:
      Attribute value of the cluster.
    """

  @abc.abstractmethod
  def write(self, endpoint_id: int, cluster: str, attribute: str,
            value: Any) -> None:
    """Writes a cluster's attribute for the given node id and endpoint.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      value: New attribute value to update the cluster with.
    """

  @abc.abstractmethod
  def start_subscription(self, endpoint_id: int, cluster: str, attribute: str,
                         min_interval_s: int, max_interval_s: int) -> None:
    """Starts a subscription to a cluster's attribute.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      min_interval_s: Server should not send a new report if less than this
        number of seconds has elapsed since the last report.
      max_interval_s: Server must send a report if this number of seconds has
        elapsed since the last report.
    """

  @abc.abstractmethod
  def stop_subscription(self) -> List[Any]:
    """Stops existing subscription and returns subscribed values.

    Returns:
      Sequence of attribute values.
    """

  @abc.abstractmethod
  def subscribe(self,
                endpoint_id: int,
                cluster: str,
                attribute: str,
                min_interval_s: int,
                max_interval_s: int,
                timeout_s: int = 10) -> List[Any]:
    """Subscribes to a cluster's attribute and blocks until timeout.

    Args:
      endpoint_id: Endpoint ID within the node to write attribute to.
      cluster: Name of the cluster to write the attribute value to (e.g. onoff).
      attribute: Name of the cluster attribute to write (e.g. on-time).
      min_interval_s: Server should not send a new report if less than this
        number of seconds has elapsed since the last report.
      max_interval_s: Server must send a report if this number of seconds has
        elapsed since the last report.
      timeout_s: Number of seconds to subscribe to this attribute for.

    Returns:
      Sequence of attribute values.
    """

  @abc.abstractmethod
  def send(self, endpoint_id: int, cluster: str, command: str,
           arguments: Sequence[Any]) -> None:
    """Sends a command to a device with the given node id and endpoint.

    Args:
      endpoint_id: Endpoint ID within the node to read attribute from.
      cluster: Name of the cluster to send the command to (e.g. onoff).
      command: Name of the command to send (e.g. toggle).
      arguments: Command arguments.
    """

  @abc.abstractmethod
  def upgrade(self, build_file: str, build_id: str,
              paa_trust_store_path: str) -> None:
    """Installs chip-tool binary to the controller device and downloads matter certs.

    Args:
      build_file: Local path to build file.
      build_id: Commit SHA the controller is built at.
      paa_trust_store_path: Path where matter certs will be downloaded.
    """

  @abc.abstractmethod
  def factory_reset(self) -> None:
    """Factory resets all settings stored on controller's device."""

  @abc.abstractmethod
  def update_certs(self,
                   device_dest: str,
                   host_dest: Optional[str] = None) -> None:
    """Downloads certs from Matter SDK and to the device.

    Downloads certs and stores in the specified path in the host. Later, copies
    them to the specified path in the device.

    Args:
      device_dest: Directory path in the device to which the certs will be
        copied from the host machine.
      host_dest: Directory path in the host machine to store the certs. If not
        specified, a temporary directory will be used and cleaned up before
        exit.
    """



#   Copyright (c) 2020 Project CHIP Authors
#   Copyright (c) 2019-2020 Google LLC.
#   All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# pylint: skip-file

"""TLV encoder and decoder module.

This module contains definitions for working with data encoded in Chip TLV format.

This is a mirror of
https://github.com/project-chip/connectedhomeip/blob/master/src/controller/python/chip/tlv/__init__.py
"""
import struct
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from enum import Enum

TLV_TYPE_SIGNED_INTEGER = 0x00
TLV_TYPE_UNSIGNED_INTEGER = 0x04
TLV_TYPE_BOOLEAN = 0x08
TLV_TYPE_FLOATING_POINT_NUMBER = 0x0A
TLV_TYPE_UTF8_STRING = 0x0C
TLV_TYPE_BYTE_STRING = 0x10
TLV_TYPE_NULL = 0x14
TLV_TYPE_STRUCTURE = 0x15
TLV_TYPE_ARRAY = 0x16
TLV_TYPE_PATH = 0x17

TLV_TAG_CONTROL_ANONYMOUS = 0x00
TLV_TAG_CONTROL_CONTEXT_SPECIFIC = 0x20
TLV_TAG_CONTROL_COMMON_PROFILE_2Bytes = 0x40
TLV_TAG_CONTROL_COMMON_PROFILE_4Bytes = 0x60
TLV_TAG_CONTROL_IMPLICIT_PROFILE_2Bytes = 0x80
TLV_TAG_CONTROL_IMPLICIT_PROFILE_4Bytes = 0xA0
TLV_TAG_CONTROL_FULLY_QUALIFIED_6Bytes = 0xC0
TLV_TAG_CONTROL_FULLY_QUALIFIED_8Bytes = 0xE0

TLVBoolean_False = TLV_TYPE_BOOLEAN
TLVBoolean_True = TLV_TYPE_BOOLEAN + 1

TLVEndOfContainer = 0x18

INT8_MIN = -128
INT16_MIN = -32768
INT32_MIN = -2147483648
INT64_MIN = -9223372036854775808

INT8_MAX = 127
INT16_MAX = 32767
INT32_MAX = 2147483647
INT64_MAX = 9223372036854775807

UINT8_MAX = 255
UINT16_MAX = 65535
UINT32_MAX = 4294967295
UINT64_MAX = 18446744073709551615

ElementTypes = {
    0x00: "Signed Integer 1-byte value",
    0x01: "Signed Integer 2-byte value",
    0x02: "Signed Integer 4-byte value",
    0x03: "Signed Integer 8-byte value",
    0x04: "Unsigned Integer 1-byte value",
    0x05: "Unsigned Integer 2-byte value",
    0x06: "Unsigned Integer 4-byte value",
    0x07: "Unsigned Integer 8-byte value",
    0x08: "Boolean False",
    0x09: "Boolean True",
    0x0A: "Floating Point 4-byte value",
    0x0B: "Floating Point 8-byte value",
    0x0C: "UTF-8 String 1-byte length",
    0x0D: "UTF-8 String 2-byte length",
    0x0E: "UTF-8 String 4-byte length",
    0x0F: "UTF-8 String 8-byte length",
    0x10: "Byte String 1-byte length",
    0x11: "Byte String 2-byte length",
    0x12: "Byte String 4-byte length",
    0x13: "Byte String 8-byte length",
    0x14: "Null",
    0x15: "Structure",
    0x16: "Array",
    0x17: "Path",
    0x18: "End of Collection",
}

TagControls = {
    0x00: "Anonymous",
    0x20: "Context 1-byte",
    0x40: "Common Profile 2-byte",
    0x60: "Common Profile 4-byte",
    0x80: "Implicit Profile 2-byte",
    0xA0: "Implicit Profile 4-byte",
    0xC0: "Fully Qualified 6-byte",
    0xE0: "Fully Qualified 8-byte",
}


class uint(int):
    '''
    NewType will not return a class until Python 3.10, as Python 3.10 is not widely used, we still need to construct a class so it can work as a type.
    '''

    def __init__(self, val: int):
        if (val < 0):
            raise TypeError(
                'expecting positive value, got negative value of %d instead' % val)


class float32(float):
    ''' A type for single precision floats distinct from the double precision 'float'
        type offered by default in Python. This type distinction is present in the Matter
        data model types so we need it here as well.
        It is backed by an ordinary float, which means there will be precision loss at the time
        the value is converted to TLV.
    '''
    pass


class TLVWriter(object):
    def __init__(self, encoding=None, implicitProfile=None):
        self._encoding = encoding if encoding is not None else bytearray()
        self._implicitProfile = implicitProfile
        self._containerStack = []

    @property
    def encoding(self):
        """The object into which encoded TLV data is written.
        By default this is a bytearray object.
        """
        return self._encoding

    @encoding.setter
    def encoding(self, val):
        self._encoding = val

    @property
    def implicitProfile(self):
        """The Chip profile id used when encoding implicit profile tags.
        Setting this value will result in an implicit profile tag being encoded
        whenever the profile of the tag to be encoded matches the specified implicit
        profile id.
        Setting this value to None (the default) disabled encoding of implicit
        profile tags.
        """
        return self._implicitProfile

    @implicitProfile.setter
    def implicitProfile(self, val):
        self._implicitProfile = val

    def put(self, tag, val):
        """Write a value in TLV format with the specified TLV tag.
        val can be a Python object which will be encoded as follows:
        - Python bools, floats and strings are encoded as their respective TLV types.
        - Python ints are encoded as unsigned TLV integers if zero or positive; signed TLV
          integers if negative.
        - None is encoded as a TLV Null.
        - bytes and bytearray objects are encoded as TVL byte strings.
        - Mapping-like objects (e.g. dict) are encoded as TLV structures.  The keys of the
          map object are expected to be tag values, as described below for the tag argument.
          Map values are encoded recursively, using the same rules as defined for the val
          argument. The encoding order of elements depends on the type of the map object.
          Elements within a dict are automatically encoded tag numerical order. Elements
          within other forms of mapping object (e.g. OrderedDict) are encoded in the
          object's natural iteration order.
        - Sequence-like objects (e.g. arrays) are written as TLV arrays. Elements within
          the array are encoded recursively, using the same rules as defined for the val
          argument.
        tag can be a small int (0-255), a tuple of two integers, or None.
        If tag is an integer, it is encoded as a TLV context-specific tag.
        If tag is a two-integer tuple, it is encoded as a TLV profile-specific tag, with
          the first integer encoded as the profile id and the second as the tag number.
        If tag is None, it is encoded as a TLV anonymous tag.
        """
        if val is None:
            self.putNull(tag)
        elif isinstance(val, Enum):
            self.putUnsignedInt(tag, val)
        elif isinstance(val, bool):
            self.putBool(tag, val)
        elif isinstance(val, uint):
            self.putUnsignedInt(tag, val)
        elif isinstance(val, int):
            self.putSignedInt(tag, val)
        elif isinstance(val, float32):
            self.putFloat(tag, val)
        elif isinstance(val, float):
            self.putDouble(tag, val)
        elif isinstance(val, str):
            self.putString(tag, val)
        elif isinstance(val, bytes) or isinstance(val, bytearray):
            self.putBytes(tag, val)
        elif isinstance(val, Mapping):
            self.startStructure(tag)
            if type(val) == dict:
                val = OrderedDict(
                    sorted(val.items(),
                           key=lambda item: tlvTagToSortKey(item[0]))
                )
            for containedTag, containedVal in val.items():
                self.put(containedTag, containedVal)
            self.endContainer()
        elif isinstance(val, Sequence):
            self.startArray(tag)
            for containedVal in val:
                self.put(None, containedVal)
            self.endContainer()
        else:
            raise ValueError("Attempt to TLV encode unsupported value")

    def putSignedInt(self, tag, val):
        """Write a value as a TLV signed integer with the specified TLV tag."""
        if val >= INT8_MIN and val <= INT8_MAX:
            format = "<b"
        elif val >= INT16_MIN and val <= INT16_MAX:
            format = "<h"
        elif val >= INT32_MIN and val <= INT32_MAX:
            format = "<l"
        elif val >= INT64_MIN and val <= INT64_MAX:
            format = "<q"
        else:
            raise ValueError("Integer value out of range")
        val = struct.pack(format, val)
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_SIGNED_INTEGER, tag, lenOfLenOrVal=len(val)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(val)

    def putUnsignedInt(self, tag, val):
        """Write a value as a TLV unsigned integer with the specified TLV tag."""
        val = self._encodeUnsignedInt(val)
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_UNSIGNED_INTEGER, tag, lenOfLenOrVal=len(val)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(val)

    def putFloat(self, tag, val):
        """Write a value as a TLV float with the specified TLV tag."""
        val = struct.pack("f", val)
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_FLOATING_POINT_NUMBER, tag, lenOfLenOrVal=len(val)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(val)

    def putDouble(self, tag, val):
        """Write a value as a TLV double with the specified TLV tag."""
        val = struct.pack("d", val)
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_FLOATING_POINT_NUMBER, tag, lenOfLenOrVal=len(val)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(val)

    def putString(self, tag, val):
        """Write a value as a TLV string with the specified TLV tag."""
        val = val.encode("utf-8")
        valLen = self._encodeUnsignedInt(len(val))
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_UTF8_STRING, tag, lenOfLenOrVal=len(valLen)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(valLen)
        self._encoding.extend(val)

    def putBytes(self, tag, val):
        """Write a value as a TLV byte string with the specified TLV tag."""
        valLen = self._encodeUnsignedInt(len(val))
        controlAndTag = self._encodeControlAndTag(
            TLV_TYPE_BYTE_STRING, tag, lenOfLenOrVal=len(valLen)
        )
        self._encoding.extend(controlAndTag)
        self._encoding.extend(valLen)
        self._encoding.extend(val)

    def putBool(self, tag, val):
        """Write a value as a TLV boolean with the specified TLV tag."""
        if val:
            type = TLVBoolean_True
        else:
            type = TLVBoolean_False
        controlAndTag = self._encodeControlAndTag(type, tag)
        self._encoding.extend(controlAndTag)

    def putNull(self, tag):
        """Write a TLV null with the specified TLV tag."""
        controlAndTag = self._encodeControlAndTag(TLV_TYPE_NULL, tag)
        self._encoding.extend(controlAndTag)

    def startContainer(self, tag, containerType):
        """Start writing a TLV container with the specified TLV tag.
        containerType can be one of TLV_TYPE_STRUCTURE, TLV_TYPE_ARRAY or
        TLV_TYPE_PATH.
        """
        self._verifyValidContainerType(containerType)
        controlAndTag = self._encodeControlAndTag(containerType, tag)
        self._encoding.extend(controlAndTag)
        self._containerStack.insert(0, containerType)

    def startStructure(self, tag):
        """Start writing a TLV structure with the specified TLV tag."""
        self.startContainer(tag, containerType=TLV_TYPE_STRUCTURE)

    def startArray(self, tag):
        """Start writing a TLV array with the specified TLV tag."""
        self.startContainer(tag, containerType=TLV_TYPE_ARRAY)

    def startPath(self, tag):
        """Start writing a TLV path with the specified TLV tag."""
        self.startContainer(tag, containerType=TLV_TYPE_PATH)

    def endContainer(self):
        """End writing the current TLV container."""
        self._containerStack.pop(0)
        controlAndTag = self._encodeControlAndTag(TLVEndOfContainer, None)
        self._encoding.extend(controlAndTag)

    def _encodeControlAndTag(self, type, tag, lenOfLenOrVal=0):
        controlByte = type
        if lenOfLenOrVal == 2:
            controlByte |= 1
        elif lenOfLenOrVal == 4:
            controlByte |= 2
        elif lenOfLenOrVal == 8:
            controlByte |= 3
        if tag is None:
            if (
                type != TLVEndOfContainer
                and len(self._containerStack) != 0
                and self._containerStack[0] == TLV_TYPE_STRUCTURE
            ):
                raise ValueError(
                    "Attempt to encode anonymous tag within TLV structure")
            controlByte |= TLV_TAG_CONTROL_ANONYMOUS
            return struct.pack("<B", controlByte)
        if isinstance(tag, int):
            if tag < 0 or tag > UINT8_MAX:
                raise ValueError(
                    "Context-specific TLV tag number out of range")
            if len(self._containerStack) == 0:
                raise ValueError(
                    "Attempt to encode context-specific TLV tag at top level"
                )
            if self._containerStack[0] == TLV_TYPE_ARRAY:
                raise ValueError(
                    "Attempt to encode context-specific tag within TLV array"
                )
            controlByte |= TLV_TAG_CONTROL_CONTEXT_SPECIFIC
            return struct.pack("<BB", controlByte, tag)
        if isinstance(tag, tuple):
            (profile, tagNum) = tag
            if not isinstance(tagNum, int):
                raise ValueError("Invalid object given for TLV tag")
            if tagNum < 0 or tagNum > UINT32_MAX:
                raise ValueError("TLV tag number out of range")
            if profile != None:
                if not isinstance(profile, int):
                    raise ValueError("Invalid object given for TLV profile id")
                if profile < 0 or profile > UINT32_MAX:
                    raise ValueError("TLV profile id value out of range")
            if (
                len(self._containerStack) != 0
                and self._containerStack[0] == TLV_TYPE_ARRAY
            ):
                raise ValueError(
                    "Attempt to encode profile-specific tag within TLV array"
                )
            if profile is None or profile == self._implicitProfile:
                if tagNum <= UINT16_MAX:
                    controlByte |= TLV_TAG_CONTROL_IMPLICIT_PROFILE_2Bytes
                    return struct.pack("<BH", controlByte, tagNum)
                else:
                    controlByte |= TLV_TAG_CONTROL_IMPLICIT_PROFILE_4Bytes
                    return struct.pack("<BL", controlByte, tagNum)
            elif profile == 0:
                if tagNum <= UINT16_MAX:
                    controlByte |= TLV_TAG_CONTROL_COMMON_PROFILE_2Bytes
                    return struct.pack("<BH", controlByte, tagNum)
                else:
                    controlByte |= TLV_TAG_CONTROL_COMMON_PROFILE_4Bytes
                    return struct.pack("<BL", controlByte, tagNum)
            else:
                vendorId = (profile >> 16) & 0xFFFF
                profileNum = (profile >> 0) & 0xFFFF
                if tagNum <= UINT16_MAX:
                    controlByte |= TLV_TAG_CONTROL_FULLY_QUALIFIED_6Bytes
                    return struct.pack("<BHHH", controlByte, vendorId, profileNum, tagNum)
                else:
                    controlByte |= TLV_TAG_CONTROL_FULLY_QUALIFIED_8Bytes
                    return struct.pack("<BHHL", controlByte, vendorId, profileNum, profile, tagNum)
        raise ValueError("Invalid object given for TLV tag")

    @staticmethod
    def _encodeUnsignedInt(val):
        if val < 0:
            raise ValueError("Integer value out of range")
        if val <= UINT8_MAX:
            format = "<B"
        elif val <= UINT16_MAX:
            format = "<H"
        elif val <= UINT32_MAX:
            format = "<L"
        elif val <= UINT64_MAX:
            format = "<Q"
        else:
            raise ValueError("Integer value out of range")
        return struct.pack(format, val)

    @staticmethod
    def _verifyValidContainerType(containerType):
        if (
            containerType != TLV_TYPE_STRUCTURE
            and containerType != TLV_TYPE_ARRAY
            and containerType != TLV_TYPE_PATH
        ):
            raise ValueError("Invalid TLV container type")


class TLVReader(object):
    def __init__(self, tlv):
        self._tlv = tlv
        self._bytesRead = 0
        self._decodings = []

    @property
    def decoding(self):
        return self._decodings

    def get(self):
        """Get the dictionary representation of tlv data"""
        out = {}
        self._get(self._tlv, self._decodings, out)
        return out

    def _decodeControlByte(self, tlv, decoding):
        (controlByte,) = struct.unpack(
            "<B", tlv[self._bytesRead: self._bytesRead + 1])
        controlTypeIndex = controlByte & 0xE0
        decoding["tagControl"] = TagControls[controlTypeIndex]
        elementtypeIndex = controlByte & 0x1F
        decoding["type"] = ElementTypes[elementtypeIndex]
        self._bytesRead += 1

    def _decodeControlAndTag(self, tlv, decoding):
        """The control byte specifies the type of a TLV element and how its tag, length and value fields are encoded.
        The control byte consists of two subfields: an element type field which occupies the lower 5 bits,
        and a tag control field which occupies the upper 3 bits. The element type field encodes the element’s type
        as well as how the corresponding length and value fields are encoded.  In the case of Booleans and the
        null value, the element type field also encodes the value itself."""

        self._decodeControlByte(tlv, decoding)

        if decoding["tagControl"] == "Anonymous":
            decoding["tag"] = None
            decoding["tagLen"] = 0
        elif decoding["tagControl"] == "Context 1-byte":
            (decoding["tag"],) = struct.unpack(
                "<B", tlv[self._bytesRead: self._bytesRead + 1]
            )
            decoding["tagLen"] = 1
            self._bytesRead += 1
        elif decoding["tagControl"] == "Common Profile 2-byte":
            profile = 0
            (tag,) = struct.unpack(
                "<H", tlv[self._bytesRead: self._bytesRead + 2])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 2
            self._bytesRead += 2
        elif decoding["tagControl"] == "Common Profile 4-byte":
            profile = 0
            (tag,) = struct.unpack(
                "<L", tlv[self._bytesRead: self._bytesRead + 4])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 4
            self._bytesRead += 4
        elif decoding["tagControl"] == "Implicit Profile 2-byte":
            profile = None
            (tag,) = struct.unpack(
                "<H", tlv[self._bytesRead: self._bytesRead + 2])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 2
            self._bytesRead += 2
        elif decoding["tagControl"] == "Implicit Profile 4-byte":
            profile = None
            (tag,) = struct.unpack(
                "<L", tlv[self._bytesRead: self._bytesRead + 4])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 4
            self._bytesRead += 4
        elif decoding["tagControl"] == "Fully Qualified 6-byte":
            (vendorId, profileNum) = struct.unpack(
                "<HH", tlv[self._bytesRead: self._bytesRead + 4])
            profile = (vendorId << 16) | profileNum
            (tag,) = struct.unpack(
                "<H", tlv[self._bytesRead + 4: self._bytesRead + 6])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 2
            self._bytesRead += 6
        elif decoding["tagControl"] == "Fully Qualified 8-byte":
            (vendorId, profileNum) = struct.unpack(
                "<HH", tlv[self._bytesRead: self._bytesRead + 4])
            profile = (vendorId << 16) | profileNum
            (tag,) = struct.unpack(
                "<L", tlv[self._bytesRead + 4: self._bytesRead + 8])
            decoding["profileTag"] = (profile, tag)
            decoding["tagLen"] = 4
            self._bytesRead += 8

    def _decodeStrLength(self, tlv, decoding):
        """UTF-8 or Byte StringLength fields are encoded in 0, 1, 2 or 4 byte widths, as specified by
        the element type field. If the element type needs a length field grab the next bytes as length"""
        if "length" in decoding["type"]:
            if "1-byte" in decoding["type"]:
                (decoding["strDataLen"],) = struct.unpack(
                    "<B", tlv[self._bytesRead: self._bytesRead + 1]
                )
                decoding["strDataLenLen"] = 1
                self._bytesRead += 1
            elif "2-byte" in decoding["type"]:
                (decoding["strDataLen"],) = struct.unpack(
                    "<H", tlv[self._bytesRead: self._bytesRead + 2]
                )
                decoding["strDataLenLen"] = 2
                self._bytesRead += 2
            elif "4-byte" in decoding["type"]:
                (decoding["strDataLen"],) = struct.unpack(
                    "<L", tlv[self._bytesRead: self._bytesRead + 4]
                )
                decoding["strDataLenLen"] = 4
                self._bytesRead += 4
            elif "8-byte" in decoding["type"]:
                (decoding["strDataLen"],) = struct.unpack(
                    "<Q", tlv[self._bytesRead: self._bytesRead + 8]
                )
                decoding["strDataLenLen"] = 8
                self._bytesRead += 8
        else:
            decoding["strDataLen"] = 0
            decoding["strDataLenLen"] = 0

    def _decodeVal(self, tlv, decoding):
        """decode primitive tlv value to the corresponding python value, tlv array and path are decoded as
        python list, tlv structure is decoded as python dictionary"""
        if decoding["type"] == "Structure":
            decoding["value"] = {}
            decoding["Structure"] = []
            self._get(tlv, decoding["Structure"], decoding["value"])
        elif decoding["type"] == "Array":
            decoding["value"] = []
            decoding["Array"] = []
            self._get(tlv, decoding["Array"], decoding["value"])
        elif decoding["type"] == "Path":
            decoding["value"] = []
            decoding["Path"] = []
            self._get(tlv, decoding["Path"], decoding["value"])
        elif decoding["type"] == "Null":
            decoding["value"] = None
        elif decoding["type"] == "End of Collection":
            decoding["value"] = None
        elif decoding["type"] == "Boolean True":
            decoding["value"] = True
        elif decoding["type"] == "Boolean False":
            decoding["value"] = False
        elif decoding["type"] == "Unsigned Integer 1-byte value":
            (decoding["value"],) = struct.unpack(
                "<B", tlv[self._bytesRead: self._bytesRead + 1]
            )
            decoding["value"] = uint(decoding["value"])
            self._bytesRead += 1
        elif decoding["type"] == "Signed Integer 1-byte value":
            (decoding["value"],) = struct.unpack(
                "<b", tlv[self._bytesRead: self._bytesRead + 1]
            )
            self._bytesRead += 1
        elif decoding["type"] == "Unsigned Integer 2-byte value":
            (decoding["value"],) = struct.unpack(
                "<H", tlv[self._bytesRead: self._bytesRead + 2]
            )
            decoding["value"] = uint(decoding["value"])
            self._bytesRead += 2
        elif decoding["type"] == "Signed Integer 2-byte value":
            (decoding["value"],) = struct.unpack(
                "<h", tlv[self._bytesRead: self._bytesRead + 2]
            )
            self._bytesRead += 2
        elif decoding["type"] == "Unsigned Integer 4-byte value":
            (decoding["value"],) = struct.unpack(
                "<L", tlv[self._bytesRead: self._bytesRead + 4]
            )
            decoding["value"] = uint(decoding["value"])
            self._bytesRead += 4
        elif decoding["type"] == "Signed Integer 4-byte value":
            (decoding["value"],) = struct.unpack(
                "<l", tlv[self._bytesRead: self._bytesRead + 4]
            )
            self._bytesRead += 4
        elif decoding["type"] == "Unsigned Integer 8-byte value":
            (decoding["value"],) = struct.unpack(
                "<Q", tlv[self._bytesRead: self._bytesRead + 8]
            )
            decoding["value"] = uint(decoding["value"])
            self._bytesRead += 8
        elif decoding["type"] == "Signed Integer 8-byte value":
            (decoding["value"],) = struct.unpack(
                "<q", tlv[self._bytesRead: self._bytesRead + 8]
            )
            self._bytesRead += 8
        elif decoding["type"] == "Floating Point 4-byte value":
            (decoding["value"],) = struct.unpack(
                "<f", tlv[self._bytesRead: self._bytesRead + 4]
            )
            decoding["value"] = float32(decoding["value"])
            self._bytesRead += 4
        elif decoding["type"] == "Floating Point 8-byte value":
            (decoding["value"],) = struct.unpack(
                "<d", tlv[self._bytesRead: self._bytesRead + 8]
            )
            self._bytesRead += 8
        elif "UTF-8 String" in decoding["type"]:
            (val,) = struct.unpack(
                "<%ds" % decoding["strDataLen"],
                tlv[self._bytesRead: self._bytesRead + decoding["strDataLen"]],
            )
            try:
                decoding["value"] = str(val, "utf-8")
            except Exception as ex:
                decoding["value"] = val
            self._bytesRead += decoding["strDataLen"]
        elif "Byte String" in decoding["type"]:
            (val,) = struct.unpack(
                "<%ds" % decoding["strDataLen"],
                tlv[self._bytesRead: self._bytesRead + decoding["strDataLen"]],
            )

            decoding["value"] = val
            self._bytesRead += decoding["strDataLen"]
        else:
            raise ValueError("Attempt to decode unsupported TLV type")

    def _get(self, tlv, decodings, out):
        endOfEncoding = False

        while len(tlv[self._bytesRead:]) > 0 and endOfEncoding == False:
            decoding = {}
            self._decodeControlAndTag(tlv, decoding)
            self._decodeStrLength(tlv, decoding)
            self._decodeVal(tlv, decoding)
            decodings.append(decoding)

            if decoding["type"] == "End of Collection":
                endOfEncoding = True
            else:
                if "profileTag" in list(decoding.keys()):
                    out[decoding["profileTag"]] = decoding["value"]
                elif "tag" in list(decoding.keys()):
                    if isinstance(out, Mapping):
                        tag = decoding["tag"] if decoding["tag"] is not None else "Any"
                        out[tag] = decoding["value"]
                    else:
                        out.append(decoding["value"])
                else:
                    raise ValueError("Attempt to decode unsupported TLV tag")


def tlvTagToSortKey(tag):
    if tag is None:
        return -1
    if isinstance(tag, int):
        majorOrder = 0
    elif isinstance(tag, tuple):
        (profileId, tag) = tag
        if profileId is None:
            majorOrder = 1
        else:
            majorOrder = profileId + 2
    else:
        raise ValueError("Invalid TLV tag")
    return (majorOrder << 32) + tag



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""chip-tool implementation of the Matter On/Off cluster capability.
"""
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.matter_clusters.interfaces import on_off_base

logger = gdm_logger.get_logger()

_CLUSTER_NAME = 'onoff'


class OnOffClusterChipTool(on_off_base.OnOffClusterBase):
  """chip-tool based Matter On/Off cluster capability."""

  @decorators.CapabilityLogDecorator(logger)
  def on(self, verify: bool = True) -> None:
    """The On command to turn on the Matter device."""
    self._onoff_command('on', True, verify)

  @decorators.CapabilityLogDecorator(logger)
  def off(self, verify: bool = True) -> None:
    """The Off command to turn off the Matter device."""
    self._onoff_command('off', False, verify)

  @decorators.CapabilityLogDecorator(logger)
  def toggle(self, verify: bool = True) -> None:
    """The Toggle command to toggle the OnOff state of the Matter device."""
    self._onoff_command('toggle', not self.onoff, verify)

  @decorators.DynamicProperty
  def onoff(self) -> bool:
    """The OnOff attribute."""
    return self._read(self._endpoint_id, _CLUSTER_NAME, 'on-off')

  def _onoff_command(self, command: str, on: bool, verify: bool = True) -> None:
    """Turn on/off the device.

    Args:
      command: Name of the command to send to the device via chip-tool.
      on: Turns on the device if true, turn off the device otherwise.
      verify: If true, verifies the device configurations before returning.

    Raises:
      DeviceError: when the device does not transition to the appropriate
      OnOff state.
    """
    self._send(self._endpoint_id, _CLUSTER_NAME, command, [])

    if verify and on is not self.onoff:
      action = 'on' if on else 'off'
      raise errors.DeviceError(
          f'Device {self._device_name} did not turn {action}.')
